#cython: language_level=3
#cython: boundscheck=False
#cython: cdivision=True

import string
import numpy as np
from neoloop.assembly import complexSV
from scipy import optimize, sparse
from neoloop.pastis import possion_model

def MDS_obj(X, distances):

    X = X.reshape(-1, 3)
    dis = np.sqrt(((X[distances.row] - X[distances.col])**2).sum(axis=1))

    return ((dis - distances.data)**2 / distances.data**2).sum()

def MDS_gradient(X, distances):

    X = X.reshape(-1, 3)
    dis = np.sqrt(((X[distances.row] - X[distances.col])**2).sum(axis=1))

    grad = 2 * ((dis - distances.data) / dis /
                distances.data**2)[:, np.newaxis] * (
        X[distances.row] - X[distances.col])
    grad_ = np.zeros(X.shape)

    for i in range(X.shape[0]):
        grad_[i] += grad[distances.row == i].sum(axis=0)
        grad_[i] -= grad[distances.col == i].sum(axis=0)

    X = X.flatten()

    return grad_.flatten()

def poisson_obj(X, counts, alpha=-3., beta=1., bias=None):

    if bias is None:
        bias = np.ones(counts.shape[0])
    bias = bias.flatten()
    
    dis = np.sqrt(((X[counts.row] - X[counts.col])**2).sum(axis=1))
    fdis = bias[counts.row] * bias[counts.col] * beta * dis ** alpha

    obj = fdis.sum() - (counts.data * np.log(fdis)).sum()
    if np.isnan(obj):
        raise ValueError("Objective function is nan")

    return obj

def poisson_gradient(X, counts, alpha=-3., beta=1., bias=None):

    if bias is None:
        bias = np.ones(counts.shape[0])
    bias = bias.flatten()

    dis = np.sqrt(((X[counts.row] - X[counts.col])**2).sum(axis=1))
    fdis = bias[counts.row] * bias[counts.col] * beta * dis ** alpha

    diff = X[counts.row] - X[counts.col]

    grad = - ((counts.data / fdis - 1) * fdis * alpha /
              (dis ** 2))[:, np.newaxis] * diff

    grad_ = np.zeros(X.shape)

    for i in range(X.shape[0]):
        grad_[i] += grad[counts.row == i].sum(axis=0)
        grad_[i] -= grad[counts.col == i].sum(axis=0)

    return grad_


def eval_f(x, user_data=None):

    n, counts, alpha, beta, bias = user_data
    x = x.reshape((n, 3))
    obj = poisson_obj(x, counts, alpha=alpha, beta=beta, bias=bias)
    x = x.flatten()

    return obj


def eval_grad_f(x, user_data=None):

    n, counts, alpha, beta, bias = user_data
    x = x.reshape((n, 3))
    grad = poisson_gradient(x, counts, alpha=alpha, beta=beta, bias=bias)
    x = x.flatten()

    return grad.flatten()


class PMCore(complexSV):

    chain_identifiers = dict(zip(range(26), string.ascii_uppercase))

    def __init__(self, clr, candidate, protocol='insitu', span=None, slopes={}, col='sweight',
        alpha=-3., beta=1., max_iter=10000, max_iter_outer_loop=5,
        seed=0):

        # reconstruct matrix
        if not col in ['weight', 'sweight']:
            if not span is None:
                wk = complexSV(clr, candidate, span=span, flexible=False, col=False,
                               protocol=protocol, slopes=slopes)
            else:
                wk = complexSV(clr, candidate, col=False, protocol=protocol, slopes=slopes)
            wk.reorganize_matrix()
            counts = wk.fusion_matrix
            x, y = counts.nonzero()
            counts[y, x] = counts[x, y]
        else:
            if not span is None:
                wk = complexSV(clr, candidate, span=span, flexible=False, col=col, protocol=protocol,
                               slopes=slopes)
                _wk = complexSV(clr, candidate, span=span, flexible=False, col=False, protocol=protocol,
                                slopes=slopes)
            else:
                wk = complexSV(clr, candidate, col=col, protocol=protocol, slopes=slopes)
                _wk = complexSV(clr, candidate, col=False, protocol=protocol, slopes=slopes)
            wk.reorganize_matrix()
            wk.correct_heterozygosity()
            counts = wk.hcm

            _wk.reorganize_matrix()
            raw = _wk.fusion_matrix # this is an upper triangular matrix
            marg = raw.sum(0) + raw.sum(1) - raw.diagonal(0)
            scale = marg[marg > 0].mean()

            counts = counts * scale
        
        counts = sparse.coo_matrix(counts)
        counts.setdiag(0)
        counts.eliminate_zeros()
        self.fusion_matrix = counts

        # PM2 parameters
        self.alpha = alpha
        self.beta = beta
        self.max_iter = max_iter
        self.max_iter_outer_loop = max_iter_outer_loop
        self.random_state = np.random.RandomState(seed=seed)

        # mark different chains (when write a PDB file)
        self.chains = {}
        for i, r in enumerate(wk.index):
            for ri in range(r[0], r[1]):
                self.chains[ri] = self.chain_identifiers[i]

    
    def _compute_wish_distances(self):

        if self.beta == 0:
            raise ValueError("beta cannot be equal to 0.")

        wish_distances = self.fusion_matrix / self.beta
        wish_distances.data[wish_distances.data != 0] **= 1. / self.alpha

        return wish_distances

    
    def _estimate_X(self):

        n = self.fusion_matrix.shape[0]
        ini = 1 - 2 * self.random_state.rand(n * 3)
        distances = self._compute_wish_distances()
        results = optimize.fmin_l_bfgs_b(
            MDS_obj, ini.flatten(),
            MDS_gradient,
            (distances,),
            iprint=0,
            factr=1e12,
            maxiter=self.max_iter
        )

        return results[0].reshape(-1, 3)

    
    def estimate_X(self, ini):

        n = self.fusion_matrix.shape[0]

        ini = np.array(ini)

        data = (n, self.fusion_matrix, self.alpha_, self.beta_, None)

        results = optimize.fmin_l_bfgs_b(
            eval_f,
            ini.flatten(),
            eval_grad_f,
            (data, ),
            iprint=0,
            maxiter=self.max_iter,
            )

        results = results[0].reshape(-1, 3)

        return results
    
    def fit(self):

        X = self._estimate_X() # initialization strategy: MDS2
        self.alpha_ = self.alpha
        self.beta_ = self.beta
        for _ in range(self.max_iter_outer_loop):
            self.alpha_, self.beta_ = possion_model.estimate_alpha_beta(
                self.fusion_matrix,
                X, bias=None, ini=[self.alpha_, self.beta_],
                random_state=self.random_state
            )
            X_ = self.estimate_X(ini=X)
        
        self.coords = X_
    
    def output_pdb(self, filename, ref_dis=1.1):

        from scipy.spatial.distance import euclidean

        X = self.coords
        q_dis = np.mean([euclidean(X[i+1], X[i]) for i in range(len(X)-1)])
        X = X * ref_dis / q_dis
        with open(filename, 'w') as out:
            for idx in range(X.shape[0]):
                # 1 - 6, Record name
                line = '{0:6s}'.format('ATOM')
                # 7 - 11, Atom serial number
                line += '{0:>5d}'.format(idx+1)
                line += ' '
                # 13 - 16, Atom name
                line += '{0:<4s}'.format('CA')
                # 17, alternate location indicator
                line += ' '
                # 18 - 20, Residue name
                line += 'GLY'
                line += ' '
                # 22, chain identifier
                line += '{0}'.format(self.chains[idx])
                # 23 - 26, residue sequence number
                line += '{0:>4d}'.format(idx+1)
                # 27, insertion code
                line += ' '
                # 28 - 30, blank
                line += '   '
                # 31 - 38
                line += '{0:>8.3f}'.format(X[idx][0])
                # 39 - 46
                line += '{0:>8.3f}'.format(X[idx][1])
                # 47 - 54
                line += '{0:>8.3f}'.format(X[idx][2])
                # 55 - 60, Occupancy
                line += '{0:>6.2f}'.format(1.0)
                # 61 - 66, Temperature factor
                line += '{0:>6.2f}'.format(75.0)
                # 67 - 78, blank
                line += '            '
                # 79 - 80, charge on the atom
                line += ' C\n'
                out.write(line)
            # CONECT section
            for i in range(1, X.shape[0]):
                # 1 - 6, Record name
                line = '{0:6s}'.format('CONECT')
                # 7 - 11, atom serial number
                line += '{0:>5d}'.format(i)
                # 12 - 16, serial number of bonded atom
                line += '{0:>5d}\n'.format(i+1)
                out.write(line)


#cython: language_level=3
#cython: boundscheck=False
#cython: cdivision=True

import pyBigWig
import numpy as np
from coolbox.plots.track.bigwig import PlotBigWig
from coolbox.plots.track.base import TrackPlot

def plot_y_axis(y_ax, ymin, ymax, size):
    """
    Plot the scale of the y axis with respect to the plot_axis
    plot something that looks like this:
    ymax ┐
            │
            │
    ymin ┘
    Parameters
    ----------
    y_ax : matplotlib.axes.Axes
        Axis to use to plot the scale
    """
    def value_to_str(value):
        if value % 1 == 0:
            str_value = str(int(value))
        else:
            if value < 0.01:
                str_value = "{:.4f}".format(value)
            else:
                str_value = "{:.2f}".format(value)
        return str_value

    ymax_str = value_to_str(ymax)
    ymin_str = value_to_str(ymin)
    x_pos = [0, 0.5, 0.5, 0]
    y_pos = [0.01, 0.01, 0.99, 0.99]
    y_ax.plot(x_pos, y_pos, color='black', linewidth=1, transform=y_ax.transAxes)
    y_ax.text(-0.2, -0.01, ymin_str, verticalalignment='bottom', horizontalalignment='right',
                transform=y_ax.transAxes, fontsize=size)
    y_ax.text(-0.2, 1, ymax_str, verticalalignment='top', horizontalalignment='right',
                transform=y_ax.transAxes, fontsize=size)
    y_ax.patch.set_visible(False)

class SigTrack(object):
    """
    Map a reference signal track in .bigwig to a local assembly.
    """
    def __init__(self, file_, bounds, orients, res=10000, nBins=500, multiply=1,
        default_chrom='chrN'):

        self.factor = int(np.ceil(nBins / (bounds[-1][0]+1)))
        self.arr = np.zeros((bounds[-1][0]+1) * self.factor)
        self.multiply = multiply

        blocks = []
        for i in range(0, len(bounds), 2):
            b1 = bounds[i][1]
            b2 = bounds[i+1][1]
            if b1[1] > b2[1]:
                b1, b2 = b2, b1
            o = orients[i//2]
            si = bounds[i][0] * self.factor
            ei = (bounds[i+1][0] + 1) * self.factor
            blocks.append((si, ei, b1[0], b1[1], b2[1], o))
        self.blocks = blocks
        self.res = res
        self.chromsizes = {default_chrom: bounds[-1][0]*res}

        self.bw = pyBigWig.open(file_)


    def chroms(self):

        return self.chromsizes
    
    def stats(self, chrom, start, end, nBins):
        '''
        The input parameters doesn't mean too much, just to make a consistent interface with CoolBox.
        '''
        for si, ei, c, s, e, o in self.blocks:
            part = np.array(self.bw.stats(c, s, e, nBins=ei-si)).astype(float)
            if o == '+':
                self.arr[si:ei] = part
            else:
                self.arr[si:ei] = part[::-1]
        
        self.arr = self.arr * self.multiply

        return self.arr

class plotSignal(PlotBigWig):

    def __init__(self, bw, **kwargs):

        properties_dict = {
            'file': 'inMemory.bw',
            'signal': bw,
            'height': 3,
            'color': '#dfccde',
            'style': 'fill',
        }

        properties_dict.update(kwargs)
        properties_dict['type'] = properties_dict['style']

        self.properties = properties_dict
        self.bw = bw
    
    def plot_y_axis(self, y_ax, size):
        """
        Plot the scale of the y axis with respect to the plot_axis
        plot something that looks like this:
        ymax ┐
             │
             │
        ymin ┘
        Parameters
        ----------
        y_ax : matplotlib.axes.Axes
            Axis to use to plot the scale
        """
        plot_axis = self.ax
        ymin, ymax = plot_axis.get_ylim()

        plot_y_axis(y_ax, ymin, ymax, size)
    
    def plot_range_text(self, ymin, ymax, fontsize):

        genome_range = self.genome_range
        ydelta = ymax - ymin

        # set min max
        format_lim = lambda lim: int(lim) if float(lim) %1 == 0 else "{:.2f}".format(lim)
        ymax_print = format_lim(ymax)
        ymin_print = format_lim(ymin)
        small_x = 0.01 * genome_range.length
        # by default show the data range
        self.ax.text(genome_range.start - small_x, ymax - ydelta * 0.2,
                     "[ {} ~ {} ]".format(ymin_print, ymax_print),
                     horizontalalignment='left',
                     verticalalignment='top',
                     fontsize=fontsize)


        


"""
Created on Wed Jun 27 18:48:42 2018

@author: XiaoTao Wang
"""

import pickle, logging, neoloop, os, joblib, math
import numpy as np
from scipy import sparse, stats
from scipy.stats import poisson
from sklearn.linear_model import HuberRegressor, LinearRegression
from sklearn.isotonic import IsotonicRegression
from neoloop.util import find_chrom_pre, map_coordinates
from scipy.stats import spearmanr

log = logging.getLogger(__name__)

def check_increasing(x, y):
    """Determine whether y is monotonically correlated with x.
    y is found increasing or decreasing with respect to x based on a Spearman
    correlation test.
    Parameters
    ----------
    x : array-like of shape (n_samples,)
            Training data.
    y : array-like of shape (n_samples,)
        Training target.
    Returns
    -------
    increasing_bool : boolean
        Whether the relationship is increasing or decreasing.
    Notes
    -----
    The Spearman correlation coefficient is estimated from the data, and the
    sign of the resulting estimate is used as the result.
    In the event that the 95% confidence interval based on Fisher transform
    spans zero, a warning is raised.
    References
    ----------
    Fisher transformation. Wikipedia.
    https://en.wikipedia.org/wiki/Fisher_transformation
    """

    # Calculate Spearman rho estimate and set return accordingly.
    rho, _ = spearmanr(x, y)
    warning = False
    increasing_bool = rho >= 0

    # Run Fisher transform to get the rho CI, but handle rho=+/-1
    if rho not in [-1.0, 1.0] and len(x) > 3:
        F = 0.5 * math.log((1. + rho) / (1. - rho))
        F_se = 1 / math.sqrt(len(x) - 3)

        # Use a 95% CI, i.e., +/-1.96 S.E.
        # https://en.wikipedia.org/wiki/Fisher_transformation
        rho_0 = math.tanh(F - 1.96 * F_se)
        rho_1 = math.tanh(F + 1.96 * F_se)

        # Warn if the CI spans zero.
        if np.sign(rho_0) != np.sign(rho_1):
            warning = True

    return warning, increasing_bool


class Fusion(object):

    def __init__(self, clr, c1, c2, p1, p2, s1, s2, sv_type, span=5000000,
        col='sweight', trim=True, protocol='insitu'):

        self.clr = clr
        self.p1 = p1
        self.p2 = p2
        self.s1 = s1
        self.s2 = s2
        self.res = clr.binsize
        pre = find_chrom_pre(list(self.clr.chromnames))
        self.c1 = pre + c1.lstrip('chr')
        self.c2 = pre + c2.lstrip('chr')
        self.chromsize1 = self.clr.chromsizes[self.c1]
        self.chromsize2 = self.clr.chromsizes[self.c2]
        self.balance_type = col
        self.protocol = protocol

        self.name = ','.join(map(str,
                                [sv_type, c1.lstrip('chr'), p1, s1, c2.lstrip('chr'), p2, s2]))

        self.get_matrix(span, col, trim)
        self.load_expected()
    
    def load_expected(self):

        data_folder = os.path.join(os.path.split(neoloop.__file__)[0], 'data')
        paths = {'dilution':{
            10000: os.path.join(data_folder, 'gm.dilution.expected.10k.pkl'),
            20000: os.path.join(data_folder, 'gm.dilution.expected.20k.pkl'),
            25000: os.path.join(data_folder, 'gm.dilution.expected.25k.pkl'),
            40000: os.path.join(data_folder, 'gm.dilution.expected.40k.pkl'),
            50000: os.path.join(data_folder, 'gm.dilution.expected.50k.pkl')
        },
        'insitu':{
            5000: os.path.join(data_folder, 'gm.insitu.expected.5k.pkl'),
            10000: os.path.join(data_folder, 'gm.insitu.expected.10k.pkl'),
            20000: os.path.join(data_folder, 'gm.insitu.expected.20k.pkl'),
            25000: os.path.join(data_folder, 'gm.insitu.expected.25k.pkl')
        }
        }
        
        self.expected = {}
        if self.protocol in paths:
            if self.res in paths[self.protocol]:
                with open(paths[self.protocol][self.res], 'rb') as source:
                    self.expected = pickle.load(source)
         

    def get_matrix(self, span, col, trim):

        res = self.res
        chromsize1 = self.chromsize1
        chromsize2 = self.chromsize2
        p1 = self.p1 // res
        p2 = self.p2 // res
        r = span // res
        
        # 4 kind of directions
        if (self.s1 == '+') and (self.s2 == '+'):
            minx_1 = max(p1-r, 0)
            if self.c1 != self.c2:
                minx_2 = max(p2 - r, 0)
            else:
                if trim:
                    minx_2 = max(p2 - r, p1 + int((p2 - p1) / 2)) # handle short inversion ++
                else:
                    minx_2 = max(p2 - r, p1 + 1)

            k_p = [minx_1 * res, min(p1 * res + res, chromsize1),
                   min(p2 * res + res, chromsize2), minx_2 * res]
            # part 1
            M1 = self.clr.matrix(balance=col).fetch((self.c1, k_p[0], k_p[1]))
            # part 2
            M2 = self.clr.matrix(balance=col).fetch((self.c2, k_p[3], k_p[2]))
            M2 = M2[::-1,::-1]
            # part 3
            M3 = self.clr.matrix(balance=col).fetch((self.c1, k_p[0], k_p[1]),
                                                    (self.c2, k_p[3], k_p[2]))
            M3 = M3[:,::-1]
        elif (self.s1 == '+') and (self.s2 == '-'):
            minx = max(p1 - r, 0)
            maxx = min((p2 + r + 1) * res, chromsize2)
            k_p = [minx * res, min(p1 * res + res, chromsize1),
                   p2 * res, maxx]
            # part 1
            M1 = self.clr.matrix(balance=col).fetch((self.c1, k_p[0], k_p[1]))
            # part 2
            M2 = self.clr.matrix(balance=col).fetch((self.c2, k_p[2], k_p[3]))
            # part 3
            M3 = self.clr.matrix(balance=col).fetch((self.c1, k_p[0], k_p[1]),
                                                    (self.c2, k_p[2], k_p[3]))
        elif (self.s1 == '-') and (self.s2 == '-'):
            if self.c1 != self.c2:
                maxx_1 = min((p1 + r + 1) * res, chromsize1)
            else:
                if trim:
                    maxx_1 = min((p1 + r + 1) * res,
                                 (p2 - int((p2 - p1) / 2)) * res) # handle short inversion --
                else:
                    maxx_1 = min((p1 + r + 1) * res, p2 * res - res)
            maxx_2 = min((p2 + r + 1) * res, chromsize2)
            k_p = [maxx_1, p1 * res, p2 * res, maxx_2]
            # part 1
            M1 = self.clr.matrix(balance=col).fetch((self.c1, k_p[1], k_p[0]))
            M1 = M1[::-1,::-1] # - --> +
            # part 2
            M2 = self.clr.matrix(balance=col).fetch((self.c2, k_p[2], k_p[3]))
            # part 3
            M3 = self.clr.matrix(balance=col).fetch((self.c1, k_p[1], k_p[0]),
                                                    (self.c2, k_p[2], k_p[3]))
            M3 = M3[::-1,:] # - --> +, + --> -
        else:
            if self.c1 != self.c2:
                maxx = min((p1 + r + 1) * res, chromsize1)
                minx = max(p2 - r, 0)
                k_p = [maxx, p1 * res,
                    min(p2 * res + res, chromsize2), minx * res] # no intra SV is -+ ?
            else:
                maxx_1 = min(p1 + r + 1, p2 - int((p2 - p1) / 2) - 1)
                minx_2 = max(p2 - r, p1 + int((p2 - p1) / 2) + 1)
                k_p = [maxx_1 * res, p1 * res, min(p2 * res + res, chromsize2), minx_2 * res]

            # part 1
            M1 = self.clr.matrix(balance=col).fetch((self.c1, k_p[1], k_p[0]))
            M1 = M1[::-1,::-1] # - --> +
            # part 2
            M2 = self.clr.matrix(balance=col).fetch((self.c2, k_p[3], k_p[2]))
            M2 = M2[::-1,::-1] # - --> +
            # part 3
            M3 = self.clr.matrix(balance=col).fetch((self.c1, k_p[1], k_p[0]),
                                                    (self.c2, k_p[3], k_p[2]))
            M3 = M3[::-1,::-1] # - --> +, - --> +
        
        # put three parts together
        size = M1.shape[0] + M2.shape[0]
        M = np.zeros((size, size))
        M[:M1.shape[0], :M1.shape[0]] = M1
        M[M1.shape[0]:, M1.shape[0]:] = M2
        M[:M1.shape[0], M1.shape[0]:] = M3
        M = np.triu(M)
    
        p1L = M1.shape[0]
        
        if col:
            M[np.isnan(M)] = 0
        
        self.fusion_matrix = M
        self.k_p = k_p
        self.fusion_point = p1L

    
    def extract_bias(self, col='sweight'):
        
        k_p = self.k_p
        if k_p[0] < k_p[1]:
            t1 = self.clr.bins().fetch((self.c1, k_p[0], k_p[1]))[col].values
        else:
            t1 = self.clr.bins().fetch((self.c1, k_p[1], k_p[0]))[col].values
            t1 = t1[::-1]
        
        if k_p[2] < k_p[3]:
            t2 = self.clr.bins().fetch((self.c2, k_p[2], k_p[3]))[col].values
        else:
            t2 = self.clr.bins().fetch((self.c2, k_p[3], k_p[2]))[col].values
            t2 = t2[::-1]
        
        m1 = np.logical_not((t1==0) | np.isnan(t1))
        w1 = np.zeros_like(t1)
        w1[m1] = 1 / t1[m1]
        m2 = np.logical_not((t2==0) | np.isnan(t2))
        w2 = np.zeros_like(t2)
        w2[m2] = 1 / t2[m2]
    
        self.bias = np.r_[w1, w2]


    def detect_bounds(self):

        from sklearn.decomposition import PCA
        from scipy.ndimage import gaussian_filter

        n = self.fusion_matrix.shape[0]
        junc = self.fusion_point

        # Detect bounds by using the induced contacts
        new = self.fusion_matrix.copy()
        up_i = 0
        down_i = n - 1
        up_var_ratio = 0
        down_var_ratio = 0
        # locate the upstream and downstream bound independently
        inter = self.fusion_matrix[:junc, junc:]
        rowmask = inter.sum(axis=1) != 0
        colmask = inter.sum(axis=0) != 0
        if (rowmask.sum() >= 10) and (colmask.sum() >= 10):
            # row, upstream
            new = inter[rowmask][:,colmask]
            corr = gaussian_filter(np.corrcoef(new, rowvar=True), sigma=1)
            try:
                pca = PCA(n_components=3, whiten=True)
                pca1 = pca.fit_transform(corr)[:,0]
                up_var_ratio = pca.explained_variance_ratio_[0]
                loc = self.locate(pca1, left_most=False)
                up_i = np.where(rowmask)[0][loc] # included
            except:
                up_i = 0
                up_var_ratio = 0

            # column, downstream
            corr = gaussian_filter(np.corrcoef(new, rowvar=False), sigma=1)
            try:
                pca = PCA(n_components=3, whiten=True)
                pca1 = pca.fit_transform(corr)[:,0]
                down_var_ratio = pca.explained_variance_ratio_[0]
                loc = self.locate(pca1, left_most=True)
                down_i = np.where(colmask)[0][loc] + junc
            except:
                down_i = n - 1
                down_var_ratio = 0
        
        self.up_i = up_i
        self.down_i = down_i
        self.up_var_ratio = up_var_ratio
        self.down_var_ratio = down_var_ratio

    
    def extend_stretches(self, arr, min_seed_len=2, max_gap=1, min_stripe_len=5):

        arr.sort()
        pieces = np.split(arr, np.where(np.diff(arr)!=1)[0]+1)
        filtered = [p for p in pieces if len(p) >= min_seed_len] # can't be empty
        stripes = []
        seed = filtered[0]
        for p in filtered[1:]:
            if p[0] - seed[-1] < (max_gap + 2):
                seed = np.r_[seed, p]
            else:
                if seed[-1] - seed[0] + 1 >= min_stripe_len:
                    stripes.append([seed[0], seed[-1]+1])
                seed = p
        
        if seed[-1] - seed[0] + 1 >= min_stripe_len:
            stripes.append([seed[0], seed[-1]+1])
        
        return stripes
    
    def locate(self, pca1, left_most=True):
        
        loci_1 = self.extend_stretches(np.where(pca1>=0)[0])
        loci_2 = self.extend_stretches(np.where(pca1<=0)[0])
        loci = loci_1 + loci_2
        loci.sort()
        if left_most:
            b = loci[0][1]
        else:
            b = loci[-1][0]
        
        return b
    
    def _extract_xy(self, valid_pixel, maxdis=2000000, mink=3):

        x, y = np.where(valid_pixel)
        M = self.fusion_matrix
        maxdis = maxdis // self.res

        dis = y - x
        local_exp = {}
        for i in range(2, M.shape[0]-1):
            if i > maxdis:
                break
            mask = dis == i
            if mask.sum() >= mink:
                local_x, local_y = x[mask], y[mask]
                tmp_arr = M[local_x, local_y]
                tmp = tmp_arr.mean()
                if tmp > 0:
                    local_exp[i] = tmp
        
        return local_exp
    
    
    def linear_regression(self, exp1, exp2, min_samples=5):

        X = []; Y = []; Xi = []
        for i in sorted(exp1):
            if i in exp2:
                Xi.append(i)
                X.append(exp2[i])
                Y.append(exp1[i])
        X = np.r_[X]
        Y = np.r_[Y]
        Xi = np.r_[Xi]

        if X.size < min_samples:
            rscore = 0
            slope = 0
            warning = False
        else:
            # clean the inputs by isotonic regression
            warning, increasing_bool = check_increasing(Xi, Y)
            IR = IsotonicRegression(increasing=increasing_bool)
            IR.fit(Xi, Y)
            Y1 = IR.predict(Xi)
            vi = np.where(np.diff(Y1) < 0)[0]
            pieces = np.split(vi, np.where(np.diff(vi)!=1)[0]+1)
            si = 0
            for i in range(len(pieces)-1):
                p1 = pieces[i]
                p2 = pieces[i+1]
                if p1.size / (p2[0] - p1[0]) > 0.5:
                    si = p1[0]
                    break
            
            if si / X.size > 0.3: # if more than 1/4 data discarded
                si = vi[0]
                if si / X.size > 0.3:
                    si = 0
            
            X = X[si:]
            Y = Y[si:]

            X = X[:, np.newaxis]
            huber = HuberRegressor().fit(X, Y)
            inlier_mask = np.logical_not(huber.outliers_)
            if inlier_mask.sum() < min_samples:
                rscore = 0
                slope = 0
            else:
                sX = X[inlier_mask]
                sY = Y[inlier_mask]
                rscore = huber.score(sX, sY)
                slope = huber.coef_[0]

        return rscore, slope, warning


    def allele_slope(self):

        # this method assumes detect_bounds has been called
        self.extract_bias(col=self.balance_type)
        shape = self.fusion_matrix.shape
        Bias_M = self.bias[:,np.newaxis] * self.bias
        valid_bias = Bias_M > 0
        p1L = self.fusion_point

        for up_i, down_i in ((self.up_i, self.down_i), (0, self.bias.size-1)):
             ## test bunch of parameters
            u_g_rscores = []
            u_g_slopes = []
            d_g_rscores = []
            d_g_slopes = []
            m_g_rscores = []
            m_g_slopes = []
            for maxdis in [400000, 1000000, 1500000]:
                for mink in [3]:
                    # middle
                    inter_mask = np.zeros(shape, dtype=bool)
                    inter_mask[up_i:p1L, p1L:down_i] = True
                    valid_pixel = inter_mask & valid_bias
                    m_exp = self._extract_xy(valid_pixel, maxdis=maxdis, mink=mink)

                    # upstream
                    inter_mask = np.zeros(shape, dtype=bool)
                    inter_mask[up_i:p1L, up_i:p1L] = True
                    valid_pixel = inter_mask & valid_bias
                    u_exp = self._extract_xy(valid_pixel, maxdis=maxdis, mink=mink)

                    # downstream
                    inter_mask = np.zeros(shape, dtype=bool)
                    inter_mask[p1L:down_i, p1L:down_i] = True
                    valid_pixel = inter_mask & valid_bias
                    d_exp = self._extract_xy(valid_pixel, maxdis=maxdis, mink=mink)

                    # upstream vs gm expected
                    r, s, w = self.linear_regression(u_exp, self.expected)
                    u_g_rscores.append(r)
                    u_g_slopes.append(s)

                    # downstream vs gm expected
                    r, s, w = self.linear_regression(d_exp, self.expected)
                    d_g_rscores.append(r)
                    d_g_slopes.append(s)

                    # middle vs gm expected
                    r, s, w = self.linear_regression(m_exp, self.expected)
                    m_g_rscores.append(r)
                    m_g_slopes.append(s)
            
            u_g_rscores = np.r_[u_g_rscores]
            d_g_rscores = np.r_[d_g_rscores]
            m_g_rscores = np.r_[m_g_rscores]
            self.u_g_s = u_g_slopes[u_g_rscores.argmax()]
            self.d_g_s = d_g_slopes[d_g_rscores.argmax()]
            self.m_g_s = m_g_slopes[m_g_rscores.argmax()]
            
            self.u_g_r = u_g_rscores.max()
            self.d_g_r = d_g_rscores.max()
            self.m_g_r = m_g_rscores.max()

            if (self.u_g_r > 0.6) and (self.d_g_r > 0.6) and (self.m_g_r > 0.6) and (self.m_g_s > 0):
                break

    
    def correct_heterozygosity(self):

        ## This method assumes allele slope has been called previously
        junc = self.fusion_point
        hcm = self.fusion_matrix.copy() # heterozygosity corrected matrix
        log.info('{0}: left slope: {1}, R^2: {2}'.format(self.name, self.u_g_s, self.u_g_r))
        log.info('{0}: right slope: {1}, R^2: {2}'.format(self.name, self.d_g_s, self.d_g_r))
        log.info('{0}, middle slope: {1}, R^2: {2}'.format(self.name, self.m_g_s, self.m_g_r))
        self.u_s = self.d_s = 0
        if (self.u_g_r > 0.6) and (self.d_g_r > 0.6):
            slope = max(self.u_g_s, self.d_g_s)
            ms = min(self.u_g_s, self.d_g_s)
            hcm[:junc, :junc] = hcm[:junc, :junc] / slope
            hcm[junc:, junc:] = hcm[junc:, junc:] / slope
            self.u_s = self.m_g_s / self.u_g_s
            self.d_s = self.m_g_s / self.d_g_s
            if (self.m_g_r > 0.6) and (self.m_g_s / ms > 0.1):
                factor = min(1/self.m_g_s, 3/slope)
                hcm[self.up_i:junc, junc:self.down_i] = hcm[self.up_i:junc, junc:self.down_i] * factor

        self.hcm = hcm
    
    def remove_gaps(self):

        # call this method after correct_heterozygosity
        M = self.hcm
        w = 5
        # output a gap-free triu matrix
        mask = M.sum(axis = 0) == 0
        index = list(np.where(mask)[0])
        convert_index = np.where(np.logical_not(mask))[0]
        dim = convert_index.size # pseudo chromosome length
        if w * 2 + 1 > dim: # if too noisy, use original matrix
            index = []
            convert_index = np.arange(len(M))
        temp = np.delete(M, index, 0)
        nM = np.delete(temp, index, 1)
        forward_map = dict(zip(np.arange(len(nM)), convert_index)) # gap-free to original

        self.gap_free = nM
        self.index_map = forward_map
    
    
class Peakachu():

    def __init__(self, matrix, lower=50000, upper=3000000, res=10000, protocol='insitu'):

        lower = lower // res
        upper = upper // res
        R, C = matrix.nonzero()
        data = matrix[R, C]
        validmask = np.isfinite(data) & (C-R+1 > lower) & (C-R < upper)
        R, C, data = R[validmask], C[validmask], data[validmask]
        self.M = sparse.csr_matrix((data, (R, C)), shape=matrix.shape)
        self.ridx, self.cidx = R, C
        self.r = res
        if (res == 10000) or (res == 5000):
            self.w = 5
        else:
            self.w = 3

        self.protocol = protocol
        self.load_models()
    
    def load_models(self):

        data_folder = os.path.join(os.path.split(neoloop.__file__)[0], 'data')
        paths = {
            'dilution':{
                10000: {
                    'total': os.path.join(data_folder, 'dilution.10K.5.pkl')
                },
                20000: {
                    'total': os.path.join(data_folder, 'dilution.20K.3.pkl')
                },
                25000: {
                    'total': os.path.join(data_folder, 'dilution.25K.3.pkl')
                },
                40000: {
                    'total': os.path.join(data_folder, 'dilution.40K.3.pkl')
                },
                50000: {
                    'total': os.path.join(data_folder, 'dilution.50K.3.pkl')
                }
            },
            'insitu': {
                10000: {
                    'ctcf': os.path.join(data_folder, 'insitu.ctcf.10k.pkl'),
                    'h3k27ac': os.path.join(data_folder, 'insitu.h3k27ac.10k.pkl')
                },
                20000: {
                    'ctcf': os.path.join(data_folder, 'insitu.ctcf.20k.pkl'),
                    'h3k27ac': os.path.join(data_folder, 'insitu.h3k27ac.20k.pkl')
                },
                25000: {
                    'ctcf': os.path.join(data_folder, 'insitu.ctcf.25k.pkl'),
                    'h3k27ac': os.path.join(data_folder, 'insitu.h3k27ac.25k.pkl')
                },
                5000: {
                    'ctcf': os.path.join(data_folder, 'insitu.h3k27ac.5k.pkl'),
                    'h3k27ac': os.path.join(data_folder, 'insitu.h3k27ac.5k.pkl'),
                }
            }
        }

        _models = paths[self.protocol][self.r]
        
        models = {w:joblib.load(_models[w]) for w in _models}
        
        self.models = models
    
    def getwindow(self, coords, w):
        '''
        Extract features from center pixels.
        '''
        fea = []
        clist = []
        for x, y in coords:
            if (x - w < 0) or (y + w + 1 > self.M.shape[0]):
                # suppose it's a upper-triangular matrix
                continue
            window = self.M[x-w:x+w+1, y-w:y+w+1].toarray()
            if np.count_nonzero(window) < window.size * 0.2:
                continue
            center = window[w, w]
            ls = window.shape[0]
            ll = np.mean(window[ls-1-ls//4:ls, :1+ls//4])
            if ll == 0:
                continue
            p2LL = center / np.mean(window[ls-1-ls//4:ls, :1+ls//4])
            ranks = stats.rankdata(window, method='ordinal')
            window = np.r_[window.ravel(), ranks, p2LL]
            fea.append(window)
            clist.append((x, y))
        
        fea = np.r_[fea]
        clist = np.r_[clist]
        
        return fea, clist
    
    def predict(self, thre, no_pool=False, min_count=2, index_map=None, chains=None):
        
        coords = [(r, c) for r, c in zip(self.ridx, self.cidx)]
        loop_list = set()
        fea, clist = self.getwindow(coords, self.w)
        if len(fea.shape) != 2:
            return list(loop_list)

        for k in self.models:
            model = self.models[k]
            probas = model.predict_proba(fea)[:, 1]
            pM = sparse.csr_matrix((probas, (clist[:,0], clist[:,1])),
                                shape=self.M.shape)
            pfilter = probas >= thre
            ri = clist[:, 0][pfilter]
            ci = clist[:, 1][pfilter]
            Donuts = {(i, j): self.M[i,j] for i, j in zip(ri, ci)} # total predictions
            if not no_pool:
                tmp = self.local_clustering(Donuts, min_count=min_count, index_map=index_map,
                                            chains=chains)
            else:
                tmp = Donuts
            for i, j in tmp:
                loop_list.add((i, j, pM[i,j]))
        
        loop_list = list(loop_list)

        return loop_list
    
    def _local_search(self, candidates, Donuts, r=20000, index_map=None,
        chains=None):

        final_list = set()
        r = max(1, r//self.r)
        for i, j in candidates:
            if (i - r < 0) or (j + r + 1 > self.M.shape[0]):
                final_list.add((i, j))
            else:
                sub = self.M[i-r:i+r+1, j-r:j+r+1].toarray()
                maxi, maxj, v = i, j, Donuts[(i, j)]
                x, y = np.where(sub > v)
                for xi, yi in zip(x, y):
                    mi, mj = i - (r - xi), j - (r - yi)
                    if index_map is None:
                        if (sub[xi, yi] > v) and (chains[mi]==chains[i]) and (chains[mj]==chains[j]):
                            v = sub[xi, yi]
                            maxi, maxj = mi, mj
                    else:
                        if (sub[xi, yi] > v) and (chains[index_map[mi]]==chains[index_map[i]]) and \
                            (chains[index_map[mj]]==chains[index_map[j]]):
                            v = sub[xi, yi]
                            maxi, maxj = mi, mj
                final_list.add((maxi, maxj))
        
        return final_list
    
    def local_clustering(self, Donuts, min_count=1, r=20000, index_map=None,
        chains=None):

        byregion = {}
        if chains is None:
            byregion[(0, 0)] = Donuts
        else:
            for i, j in Donuts:
                if index_map is None:
                    x, y = i, j
                else:
                    x, y = index_map[i], index_map[j]
                c1, c2 = chains[x], chains[y]
                if not (c1, c2) in byregion:
                    byregion[(c1, c2)] = {}
                byregion[(c1, c2)][(i, j)] = Donuts[(i, j)]
        
        coords = set()
        for k in byregion:
            Donuts = byregion[k]
            tmp = self._local_clustering(Donuts, min_count=min_count, r=r)
            precise_ = self._local_search(tmp, Donuts, r=r, index_map=index_map,
                                        chains=chains)
            coords.update(precise_)
        
        return coords


    def _local_clustering(self, Donuts, min_count=1, r=20000):

        final_list = []
        x = np.r_[[i[0] for i in Donuts]]
        y = np.r_[[i[1] for i in Donuts]]
        if x.size == 0:
            return final_list

        x_anchors = self.find_anchors(x, min_count=min_count, min_dis=r)
        y_anchors = self.find_anchors(y, min_count=min_count, min_dis=r)
        r = max(r//self.r, 1)
        visited = set()
        lookup = set(zip(x, y))
        for x_a in x_anchors:
            for y_a in y_anchors:
                sort_list = []
                for i in range(x_a[1], x_a[2]+1):
                    for j in range(y_a[1], y_a[2]+1):
                        if (i, j) in lookup:
                            sort_list.append((Donuts[(i,j)], (i,j)))
                sort_list.sort(reverse=True)
                self._cluster_core(sort_list, r, visited, final_list)
        
        sort_list = [] # out of anchor
        for i, j in zip(x, y):
            if (i,j) in visited:
                continue
            sort_list.append((Donuts[(i,j)], (i,j)))
        sort_list.sort(reverse=True)
        self._cluster_core(sort_list, r, visited, final_list)
        
        for i, j in zip(x, y):
            if (i,j) in visited:
                continue

            final_list.append((int(i), int(j)))
        
        final_list = list(set(final_list))
        
        return final_list
    
    def find_anchors(self, pos, min_count=3, min_dis=20000, wlen=200000):

        from collections import Counter
        from scipy.signal import find_peaks, peak_widths

        min_dis = max(min_dis//self.r, 1)
        wlen = min(wlen//self.r, 20)

        count = Counter(pos)
        refidx = range(min(count), max(count)+1)
        signal = np.r_[[count[i] for i in refidx]]
        summits = find_peaks(signal, height=min_count, distance=min_dis)[0]
        sorted_summits = [(signal[i],i) for i in summits]
        sorted_summits.sort(reverse=True) # sort by peak count
        
        peaks = set()
        records = {}
        for _, i in sorted_summits:
            tmp = peak_widths(signal, [i], rel_height=1, wlen=wlen)[2:4]
            li, ri = int(np.round(tmp[0][0])), int(np.round(tmp[1][0]))
            lb = refidx[li]
            rb = refidx[ri]
            if not len(peaks):
                peaks.add((refidx[i], lb, rb))
                for b in range(lb, rb+1):
                    records[b] = (refidx[i], lb, rb)
            else:
                for b in range(lb, rb+1):
                    if b in records:
                        # merge anchors
                        m_lb = min(lb, records[b][1])
                        m_rb = max(rb, records[b][2])
                        summit = records[b][0] # always the highest summit
                        peaks.remove(records[b])
                        break
                else: # loop terminates normally
                    m_lb, m_rb, summit = lb, rb, refidx[i]
                peaks.add((summit, m_lb, m_rb))
                for b in range(m_lb, m_rb+1):
                    records[b] = (summit, m_lb, m_rb)
        
        return peaks

    def _cluster_core(self, sort_list, r, visited, final_list):

        from sklearn.cluster import dbscan
        from scipy.spatial.distance import euclidean

        pos = np.r_[[i[1] for i in sort_list]]
        if len(pos) >= 2:
            _, labels = dbscan(pos, eps=r, min_samples=2)
            pool = set()
            for i, p in enumerate(sort_list):
                if p[1] in pool:
                    continue
                c = labels[i]
                if c==-1:
                    continue
                sub = pos[labels==c]
                cen = p[1]
                rad = r
                Local = [p[1]]
                ini = -1
                while len(sub):
                    out = []
                    for q in sub:
                        if tuple(q) in pool:
                            continue
                        tmp = euclidean(q, cen)
                        if tmp<=rad:
                            Local.append(tuple(q))
                        else:
                            out.append(tuple(q))
                    if len(out)==ini:
                        break
                    ini = len(out)
                    tmp = np.r_[Local]
                    # assign centroid to a certain pixel
                    cen = tuple(tmp.mean(axis=0).round().astype(int))
                    rad = np.int(np.round(max([euclidean(cen,q) for q in Local]))) + r
                    sub = np.r_[out]
                for q in Local:
                    pool.add(q)
                final_list.append((int(p[1][0]), int(p[1][1])))
            
            visited.update(pool)
        
        elif len(pos)==1:
            final_list.append((int(pos[0][0]), int(pos[0][1])))


def combine_annotations(byres):
    """
    Combine loops at different resolutions.

    """
    reslist = sorted(byres)
    pool = byres[reslist[0]]
    visited = [reslist[0]]

    for res in reslist[1:]:
        cur = byres[res]
        for l in cur:
            for pr in visited:
                count = 0
                for i in range(l[1]//pr, l[2]//pr+1):
                    for j in range(l[4]//pr, l[5]//pr+1):
                        tmp = (l[0], i*pr, i*pr+pr, l[3], j*pr, j*pr+pr)
                        if tmp in pool:
                            count += 1
                            pool[tmp].extend(cur[l])
                if count == 0:
                    pool[l] = cur[l]
                     
        visited.append(res)
    
    loop_list = []
    for k in sorted(pool):
        trace = set(pool[k])
        labels = []
        for t in trace:
            labels.append(','.join([t[0], str(t[1]), str(t[2])]))
        label = ','.join(labels)
        loop_list.append(list(map(str, k))+[label])
    
    return loop_list


import logging
import numpy as np
from neoloop.util import find_chrom_pre
from neoloop.callers import Fusion
from neoloop.util import map_coordinates, load_translocation_fil
from joblib import Parallel, delayed
import networkx as nx
from networkx.algorithms import all_shortest_paths, has_path

log = logging.getLogger(__name__)

def filterSV(clr, c1, c2, p1, p2, s1, s2, note, span, col, protocol, cutoff):

    fu = Fusion(clr, c1, c2, p1, p2, s1, s2, note, span=span, col=col, protocol=protocol, trim=False)
    fu.detect_bounds()
    fu.allele_slope()
    fu.correct_heterozygosity()
    if (fu.m_g_r > cutoff) and max(fu.u_s, fu.d_s) > 0.1:
        node = (c1, p1, s1, c2, p2, s2, note)
        Map = map_coordinates(fu.k_p, clr.binsize, c1, c2)[1]
        up = Map[fu.up_i]
        down = Map[fu.down_i]
        record = [node, (c1, up[1], fu.k_p[1]), (c2, fu.k_p[2], down[1])]
        return record
    else:
        return

def filterAssembly(clr, ID, span, col, path, protocol):

    wk = complexSV(clr, ID, span, col, protocol=protocol)
    wk.reorganize_matrix()
    wk.correct_heterozygosity()
    connectable = (len(wk.pair_binary)==wk.valid_counts) and wk.connectivity

    return connectable, path


class assembleSV(object):

    def __init__(self, clr, sv_fil, span=5000000, col='sweight', minIntra=500000,
        n_jobs=1, protocol='insitu', r_cutoff=0.6):

        self.clr = clr
        self.res = clr.binsize
        self.protocol = protocol
        self.span = span
        self.balance_type = col
        self.n_jobs = n_jobs
        # read SVs
        sv_list = load_translocation_fil(sv_fil, clr.binsize, minIntra)
        
        # filter sv by checking power-law decay of the induced interactions
        params = []
        for c1, p1, s1, c2, p2, s2, note in sv_list:
            params.append((clr, c1, c2, p1, p2, s1, s2, note, span, col, protocol, r_cutoff))
        
        results = Parallel(n_jobs=n_jobs, verbose=10)(delayed(filterSV)(*i) for i in params)
        self.queue = []
        self.lookup = {}
        for r in results:
            if not r is None:
                self.queue.append(r)
                self.lookup[r[0]] = [r[1], r[2]]
    
    def build_graph(self):

        nodes = {}
        for n, u, d in self.queue:
            k1 = (n, u, d)
            n2 = n[3:6] + n[:3] + (n[-1],)
            k2 = (n2, d, u)
            nodes[k1] = n
            nodes[k2] = n

        G = nx.DiGraph()
        for sv1 in nodes:
            for sv2 in nodes:
                if nodes[sv1] == nodes[sv2]:
                    continue
                dis = self.connect(sv1, sv2)
                if dis < np.inf:
                    G.add_weighted_edges_from([(sv1, sv2, dis)])
        
        self.graph = G
        self.nodemap = nodes
    
    def find_complexSV(self):
        
        pool = []
        nodes = self.graph.nodes()
        for n1 in nodes:
            for n2 in nodes:
                if n1 == n2:
                    continue
                pre = has_path(self.graph, n1, n2)
                if pre:
                    # do not consider edge weight
                    paths = list(all_shortest_paths(self.graph, n1, n2, weight=None))
                    for p in paths:
                        if not self._containloop(p):
                            pool.append((len(p), p))
        pool.sort(reverse=True)
        
        # so far, the candidate paths contain no self-loops, but are still redundant
        # check distance-decay for each pair of regions
        queue = [(self.clr, self._change_format(p[1]), self.span, self.balance_type, p[1], self.protocol) for p in pool]
        log.info('Filtering {0} redundant candidates ...'.format(len(queue)))
        jobs = Parallel(n_jobs=self.n_jobs, verbose=10)(delayed(filterAssembly)(*i) for i in queue)
        pre_alleles = []
        for ck, p in jobs:
            if ck:
                pre_alleles.append(p)
        
        # these assembly should exist within the same allele
        alleles = []
        for p in pre_alleles:
            for v in alleles:
                if self._issubset(p, v) or self._issubset(p, self._getreverse(v)):
                    break
            else:
                alleles.append(p)
        
        self.alleles = alleles
    
    def _containloop(self, path):

        loop_hit = []
        chain = []
        for p in path:
            chain.append(p[1])
            chain.append(p[2])
        
        for i in range(len(chain)):
            for j in range(len(chain)):
                if j - i > 1:
                    r1 = chain[i]
                    r2 = chain[j]
                    o, _ = self._check_overlap(r1, r2)
                    if o > 0.05:
                        loop_hit.append((r1, r2))
        
        return loop_hit
    
    def _getreverse(self, path):

        reverse = []
        for n, u, d in path[::-1]:
            n2 = n[3:6] + n[:3] + (n[-1],)
            k2 = (n2, d, u)
            reverse.append(k2)
        
        return reverse


    def _issubset(self, p1, p2):

        index = []
        for i in p1:
            if i in p2:
                index.append(p2.index(i))
            else:
                return False
        
        idx = np.r_[index]
        df = np.diff(idx)

        return np.all(df==1)
        
    
    def _check_overlap(self, i1, i2):

        if i1[0] != i2[0]:
            return 0, 0
        
        t1 = sorted(i1[1:])
        t2 = sorted(i2[1:])

        if (t1[1]<=t2[0]) or (t2[1]<=t1[0]):
            return 0, 0
        
        mi = t1 + t2
        mi.sort()
        OR = (mi[2] - mi[1]) / (mi[3] - mi[0]) # intersect / union

        return OR, mi[2]-mi[1]

    def connect(self, r1, r2):

        n1, u1, d1 = r1
        n2, u2, d2 = r2
        s11, s12 = n1[2], n1[5]
        s21, s22 = n2[2], n2[5]

        # check overlap between the 2nd part of sv1 and the 1st part of sv2
        ori = 0
        if s12 == '-':
            if s21 == '+':
                o, itr = self._check_overlap(d1, u2)
                if (o > ori) and (itr > 4*self.res):
                    ori = o
        else:
            if s21 == '-':
                o, itr = self._check_overlap(d1, u2)
                if (o > ori) and (itr > 4*self.res):
                    ori = o
        
        if ori > 0:
            return 1 / ori
        else:
            return np.inf
    
    def _change_format(self, path):

        names = []
        for p in path:
            tmp = ','.join(list(map(str, (p[0][-1],) + p[0][:-1])))
            names.append(tmp)
        
        b1 = path[0][1]
        b2 = path[-1][2]

        if path[0][0][2] == '+':
            t1 = [b1[0], str(min(b1[1:]))]
        else:
            t1 = [b1[0], str(max(b1[1:]))]
        
        if path[-1][0][5] == '-':
            t2 = [b2[0], str(max(b2[1:]))]
        else:
            t2 = [b2[0], str(min(b2[1:]))]
        
        t1 = ','.join(t1)
        t2 = ','.join(t2)

        line = names + [t1, t2]
        line = '\t'.join(line)
        
        return line
    
    def print(self, complexSVs, outfil):

        with open(outfil, 'w') as out:
            for path in complexSVs:
                line = self._change_format(path)
                out.write(line+'\n')
    
    def output_all(self, outfil):

        with open(outfil, 'w') as out:
            # alleles
            for i, path in enumerate(self.alleles):
                line = self._change_format(path)
                out.write('A{0}\t'.format(i)+line+'\n')
            # single SVs
            for i, sv in enumerate(self.lookup):
                path = [(sv, self.lookup[sv][0], self.lookup[sv][1])]
                line = self._change_format(path)
                out.write('C{0}\t'.format(i)+line+'\n')


class complexSV(Fusion):

    def __init__(self, clr, candidate, span=5000000, col='sweight', protocol='insitu',
        flexible=True, slopes={}):

        self.clr = clr
        self.res = clr.binsize
        self.protocol = protocol
        self.pre = find_chrom_pre(list(self.clr.chromnames))
        self.span = span
        self.flexible = flexible
        self.slopes = slopes
        
        if col in ['weight', 'sweight']:
            self.balance_type = col
        else:
            self.balance_type = False

        self.chroms = {}
        for c in self.clr.chromnames:
            self.chroms[c] = [0, self.clr.chromsizes[c]]
        
        self.parse_input(candidate)
        if len(self.sv_list)==1:
            self.tb, self.to = self.get_single_block(self.sv_list[0],
                                            left_bound=self.bounds[0],
                                            right_bound=self.bounds[1])
        else:
            self.tb = []
            self.to = []
            for i, sv in enumerate(self.sv_list):
                if i == 0:
                    intervals, directs = self.get_single_block(sv, left_bound=self.bounds[0])
                elif i == len(self.sv_list)-1:
                    intervals, directs = self.get_single_block(sv, right_bound=self.bounds[1])
                else:
                    intervals, directs = self.get_single_block(sv)
                self.tb.extend(intervals)
                self.to.extend(directs)
        
        self.load_expected()
    
    def parse_input(self, candidate):

        parse = candidate.split()
        self.sv_list = []
        for p in parse[:-2]:
            _, c1, p1, s1, c2, p2, s2 = p.split(',')
            p1, p2 = int(p1), int(p2)
            self.sv_list.append((c1, c2, p1, p2, s1, s2))

        b1 = parse[-2].split(',')
        b2 = parse[-1].split(',')
        self.bounds = [(b1[0], int(b1[1])), (b2[0], int(b2[1]))]

    
    def get_single_block(self, sv, left_bound=None, right_bound=None):

        c1, c2, p1, p2, s1, s2 = sv
        c1, c2 = self.pre+c1, self.pre+c2

        directions = [s1]
        if s2 == '+':
            directions.append('-')
        else:
            directions.append('+')
        
        intervals = []
        if s1 == '+':
            if self.flexible:
                if left_bound is None:
                    r1 = [c1, max(self.chroms[c1][0], p1 - self.span), p1]
                else:
                    r1 = [c1, left_bound[1], p1]
            else:
                r1 = [c1, max(self.chroms[c1][0], p1 - self.span), p1]
        else:
            if self.flexible:
                if left_bound is None:
                    r1 = [c1, p1, min(p1 + self.span, self.chroms[c1][1])]
                else:
                    r1 = [c1, p1, left_bound[1]]
            else:
                r1 = [c1, p1, min(p1 + self.span, self.chroms[c1][1])]
        intervals.append(r1)

        if s2 == '-':
            if self.flexible:
                if right_bound is None:
                    r2 = [c2, p2, min(p2 + self.span, self.chroms[c2][1])]
                else:
                    r2 = [c2, p2, right_bound[1]]
            else:
                r2 = [c2, p2, min(p2 + self.span, self.chroms[c2][1])]
        else:
            if self.flexible:
                if right_bound is None:
                    r2 = [c2, max(self.chroms[c2][0], p2 - self.span), p2]
                else:
                    r2 = [c2, right_bound[1], p2]
            else:
                r2 = [c2, max(self.chroms[c2][0], p2 - self.span), p2]
        intervals.append(r2)

        return intervals, directions
    
    def get_bias(self, r):

        if self.balance_type == False:
            col = 'weight'
        else:
            col = self.balance_type

        bias = self.clr.bins().fetch(r)[col].values
        bias[np.isnan(bias)] = 0

        return bias
    
    def get_matrix(self, r1, r2):

        M = self.clr.matrix(balance=self.balance_type).fetch(r1, r2)

        if self.balance_type:
            M[np.isnan(M)] = 0
        
        return M
    
    def reorganize_matrix(self):

        tb = self.tb
        to = self.to
        blocks = [tb[0]]
        orients = [to[0]]
        for i in range(1, len(tb)-1, 2):

            assert to[i]==to[i+1]
            assert tb[i][0]==tb[i+1][0]

            if to[i]=='+': # 5 --> 3
                b1 = tb[i][1]
                b2 = tb[i+1][2]
                assert b1 < b2
                blocks.append([tb[i][0], b1, b2])
                orients.append('+')
            else:
                b1 = tb[i][2]
                b2 = tb[i+1][1]
                assert b1 > b2
                blocks.append([tb[i][0], b2, b1])
                orients.append('-')
        blocks.append(tb[-1])
        orients.append(to[-1])

        # generate matrix and indexing
        index = []
        start = 0
        Map = []
        bounds = []
        for r, o in zip(blocks, orients):
            N = self.clr.bins().fetch(tuple(r))['chrom'].size
            index.append((start, start + N))
            coords = []
            for i in range(start, start + N):
                coords.append((r[0], r[1]//self.res*self.res + (i-start)*self.res))
            if o=='+':
                Map.extend(list(zip(coords, range(start, start + N))))
                bounds.append((start, (r[0], r[1]//self.res*self.res)))
                bounds.append((start + N - 1, (r[0], r[2]//self.res*self.res)))
            else:
                Map.extend(list(zip(coords[::-1], range(start, start + N))))
                bounds.append((start, (r[0], r[2]//self.res*self.res)))
                bounds.append((start + N - 1, (r[0], r[1]//self.res*self.res)))

            start = index[-1][1]

        forward = dict(Map)

        M = np.zeros((index[-1][1], index[-1][1]))
        bias_arr = np.r_[[]]
        for r1, o1, i1, j1 in zip(blocks, orients, index, range(len(blocks))):
            tmp = self.get_bias(r1)
            if o1=='+':
                bias_arr = np.r_[bias_arr, tmp]
            else:
                bias_arr = np.r_[bias_arr, tmp[::-1]]
            for r2, o2, i2, j2 in zip(blocks, orients, index, range(len(blocks))):
                if j1 > j2:
                    continue
                tmp = self.get_matrix(r1, r2)
                if (o1=='+') and (o2=='-'):
                    tmp = tmp[:,::-1]
                elif (o1=='-') and (o2=='+'):
                    tmp = tmp[::-1, :]
                elif (o1=='-') and (o2=='-'):
                    tmp = tmp[::-1, ::-1]
                M[i1[0]:i1[1], i2[0]:i2[1]] = tmp
        
        #x, y = M.nonzero()
        #M[y,x] = M[x,y]

        self.fusion_matrix = np.triu(M)
        self.Map = forward
        self.reverse = {forward[i]:i for i in forward}
        self.bounds = bounds
        self.orients = orients
        self.index = index
        self.blocks = blocks
        self.bias = bias_arr
        self.chains = {}
        for i, r in enumerate(self.index):
            for ri in range(r[0], r[1]):
                self.chains[ri] = i


    def correct_heterozygosity(self):

        Bias_M = self.bias[:,np.newaxis] * self.bias
        valid_bias = np.triu(Bias_M > 0)

        hcm = self.fusion_matrix.copy()
        slopes = {}
        exps = {}
        pair_binary = {}
        for i1, j1 in zip(self.index, range(len(self.blocks))):
            for i2, j2 in zip(self.index, range(len(self.blocks))):
                if j1 > j2:
                    continue
                inter_mask = np.zeros(hcm.shape, dtype=bool)
                inter_mask[i1[0]:i1[1], i2[0]:i2[1]] = True
                valid_pixel = inter_mask & valid_bias
                _rscores = []
                _slopes = []
                warning = []
                for maxdis in [400000, 1000000, 1500000, 2000000]:
                    local_exp = self._extract_xy(valid_pixel, mink=3, maxdis=maxdis)
                    r, s, w = self.linear_regression(local_exp, self.expected, min_samples=5)
                    warning.append(w)
                    _rscores.append(r)
                    _slopes.append(s)
                _rscores = np.r_[_rscores]
                r = _rscores.max()
                s = _slopes[_rscores.argmax()]
                    
                if any(warning):
                    pair_binary[(j1, j2)] = 0
                else:
                    pair_binary[(j1, j2)] = 1
                
                slopes[(j1, j2)] = [r, s]
                exps[(j1, j2)] = local_exp
        
        
        for i1, j1 in zip(self.index, range(len(self.blocks))):
            for i2, j2 in zip(self.index, range(len(self.blocks))):
                if j1 > j2:
                    continue
                ms = max(slopes[(j1, j1)][1], slopes[(j2, j2)][1])
                if (slopes[(j1, j2)][0] > 0.6) and (slopes[(j1, j2)][1] > 0) and (ms > 0):
                    if (j1, j2) in self.slopes:
                        hcm[i1[0]:i1[1], i2[0]:i2[1]] = hcm[i1[0]:i1[1], i2[0]:i2[1]] / self.slopes[(j1, j2)]
                    else:
                        if j1 == j2:
                            hcm[i1[0]:i1[1], i2[0]:i2[1]] = hcm[i1[0]:i1[1], i2[0]:i2[1]] / slopes[(j1, j2)][1]
                        else:
                            factor = min(1/slopes[(j1, j2)][1], 3/ms)
                            hcm[i1[0]:i1[1], i2[0]:i2[1]] = hcm[i1[0]:i1[1], i2[0]:i2[1]] * factor
        
        x, y = hcm.nonzero()
        hcm[y, x] = hcm[x, y]
        
        self.slopes_ = slopes
        self.exps = exps
        self.symm_hcm = hcm
        self.hcm = np.triu(hcm)
        valid_counts = sum([pair_binary[k] for k in pair_binary])
        self.valid_counts = valid_counts
        self.pair_binary = pair_binary
        # connectivity
        table = np.zeros((len(self.blocks), len(self.blocks)))
        for i, j in slopes:
            if slopes[i,j][0] > 0.6:
                table[i,j] = 1
                table[j,i] = 1
        connectivity = True
        for i in range(1, min(table.shape[0],3)):
            if np.any(table.diagonal(i)==0):
                connectivity = False
        self.connectivity = connectivity

    
    def index_to_coordinates(self, loop_list):

        loop_indices = []
        for i, j, v in loop_list:
            x = self.index_map[i]
            y = self.index_map[j]
            loop_indices.append((x, y, v))
        
        lines = {}
        for i, j, v in loop_indices:
            gdis = (j - i) * self.res
            loci1 = self.reverse[i]
            loci2 = self.reverse[j]
            if loci1 > loci2:
                loci1, loci2 = loci2, loci1
            key = (loci1[0], loci1[1], loci1[1]+self.res,
                   loci2[0], loci2[1], loci2[1]+self.res)
            if self.chains[i]==self.chains[j]:
                lines[key] = [gdis, 0] # normal loops
            else:
                lines[key] = [gdis, 1] # neo loops
        
        return lines


#cython: language_level=3
#cython: boundscheck=False
#cython: cdivision=True

import logging, copy
from collections import defaultdict
import numpy as np
from sklearn.mixture import GaussianMixture
from neoloop.cnv.loadcnv import binCNV
from pomegranate import NormalDistribution, HiddenMarkovModel, GeneralMixtureModel, State

logger = logging.getLogger(__name__)

def cbs_stat(x):
    '''
    Given x, Compute the subinterval x[i0:i1] with the maximal segmentation statistic t. 
    Returns t, i0, i1.

    '''
    
    x0 = x - np.mean(x)
    n = len(x0)
    y = np.cumsum(x0)
    e0, e1 = np.argmin(y), np.argmax(y)
    i0, i1 = min(e0, e1), max(e0, e1)
    s0, s1 = y[i0], y[i1]

    return (s1-s0)**2*n/(i1-i0+1)/(n+1-i1+i0), i0, i1+1

def cbs(x, shuffles=1000, p=.05):
    '''
    Given x, find the interval x[i0:i1] with maximal segmentation statistic t. Test that statistic against
    given (shuffles) number of random permutations with significance p.  Return True/False, t, i0, i1; True if
    interval is significant, false otherwise.
    '''
    max_t, max_start, max_end = cbs_stat(x)
    if max_end-max_start == len(x):
        return False, max_t, max_start, max_end
    if max_start < 5:
        max_start = 0
    if len(x)-max_end < 5:
        max_end = len(x)
    thresh_count = 0
    alpha = shuffles*p
    xt = x.copy()
    for i in range(shuffles):
        np.random.shuffle(xt)
        threshold, s0, e0 = cbs_stat(xt)
        if threshold >= max_t:
            thresh_count += 1
        if thresh_count > alpha:
            return False, max_t, max_start, max_end

    return True, max_t, max_start, max_end

def rsegment(x, start, end, L=[], shuffles=1000, p=.05):
    '''
    Recursively segment the interval x[start:end] returning a list L of pairs (i,j) where each (i,j) is a significant segment.
    '''
    threshold, t, s, e = cbs(x[start:end], shuffles=shuffles, p=p)
    if (not threshold) | (e-s < 5) | (e-s == end-start):
        L.append((start, end))
    else:
        if s > 0:
            rsegment(x, start, start+s, L)
        if e-s > 0:
            rsegment(x, start+s, start+e, L)
        if start+e < end:
            rsegment(x, start+e, end, L)

    return L

def segment(x, shuffles=1000, p=1e-5):
    '''
    Segment the array x, using significance test based on shuffles rearrangements and significance level p
    '''
    start = 0
    end = len(x)
    L = []
    rsegment(x, start, end, L, shuffles=shuffles, p=p)

    return L

def combine_segment(sig, cbs_seg, hmm_seg, bufsize=4):

    start = cbs_seg[0][0]
    pool1 = set()
    for b in cbs_seg:
        for i in range(bufsize):
            pool1.add(b[1]-i)
            pool1.add(b[1]+i)

    pool2 = set([b[1] for b in hmm_seg])
    pool = [start] + sorted(pool1 & pool2)
    seg = []
    for i in range(len(pool)-1):
        s, e = pool[i], pool[i+1]
        v = np.median(sig[s:e])
        seg.extend([v]*(e-s))
    seg = np.r_[seg]

    return seg

def impute(self, arr):

    zeros = np.where(arr==0)[0]
    nonzeros = np.where(arr!=0)[0]
    fill = arr.copy()
    for i in zeros:
        idx = nonzeros.searchsorted(i)
        if (idx == 0) or (idx == nonzeros.size):
            continue # leading or tailing 0s
        ri = nonzeros[idx]
        li = nonzeros[idx-1]
        if arr[li] < arr[ri]:
            fill[i] = arr[li]
        else:
            fill[i] = arr[ri]
    
    return fill

class HMMsegment(binCNV):

    def __init__(self, bedgraph, res, nproc=1, ploidy=2):

        binCNV.__init__(self, bedgraph, res)
        self.n_jobs = nproc
        self.ploidy = ploidy

    def segment(self, min_seg=20):

        outlines = []
        counts = defaultdict(int)
        vMap = {}
        self._original = {}
        for c in self.bin_cnv:
            logger.info('Segmenting Chromosome {0} ...'.format(c))
            sig = self.bin_cnv[c]
            sig, hmm_seg, scale = self._segment(sig) # pure HMM-based segmentation
            self._original[c] = [sig, hmm_seg]
            logger.info('Combine results from the CBS algorithm ...')
            if scale=='log':
                logsig = np.zeros_like(sig)
                logsig[sig>0] = np.log2(sig[sig>0])
                cbs_seg = segment(logsig)
            else:
                cbs_seg = segment(sig)
            seg = combine_segment(sig, cbs_seg, hmm_seg) # combine results from cbs and hmm
            
            cur = [c, 0, 1, seg[0]]
            for i in range(1, seg.size):
                v = seg[i]
                if v == cur[-1]:
                    cur[2] = cur[2] + 1
                else:
                    tmp = sig[cur[1]:cur[2]]
                    if tmp.sum() == 0:
                        cn = 0
                    else:
                        cn = int(np.round(np.median(tmp[tmp!=0])*self.ploidy))
                    cur[3] = cn
                    vMap[cn] = cn
                    counts[cn] += (cur[2]-cur[1])
                    outlines.append(cur)
                    start = cur[2]
                    cur = [c, start, start+1, v]
            
            tmp = sig[cur[1]:cur[2]]
            if tmp.sum() == 0:
                cn = 0
            else:
                cn = int(np.round(np.median(tmp[tmp!=0])*self.ploidy))
            cur[3] = cn
            vMap[cn] = cn
            counts[cn] += (cur[2]-cur[1])
            outlines.append(cur)
        
        cn_min_count = min(counts.values())
        while cn_min_count < min_seg:
            for i in counts:
                if counts[i]==cn_min_count:
                    cn = i
            neighbor = 0
            mindis = 1000 # a very large number to start
            for i in counts:
                if i == cn:
                    continue
                if np.abs(i - cn) < mindis:
                    neighbor = vMap[i]
                    mindis = np.abs(i - cn)

            tc = counts[cn] + counts[neighbor]

            for i in vMap:
                if vMap[i] == cn:
                    vMap[i] = neighbor
            
            for i in vMap:
                if vMap[i] == neighbor:
                    counts[i] = tc
            
            cn_min_count = min(counts.values())
        
        for line in outlines:
            line[-1] = vMap[line[-1]]
        
        self.outlines = []
        for i, line in enumerate(outlines):
            if not len(self.outlines):
                self.outlines.append(line)
            else:
                if line[0] != self.outlines[-1][0]:
                    self.outlines.append(line)
                else:
                    if (line[3] == 0) and (i < len(outlines)-1) and (line[0] == outlines[i+1][0]) and (line[2] - line[1] < 5):
                        self.outlines[-1][2] = line[2]
                    else:
                        if line[3] != self.outlines[-1][3]:
                            self.outlines.append(line)
                        else:
                            self.outlines[-1][2] = line[2]
    
    def output(self, fil, pre='chr'):

        with open(fil, 'w') as out:
            for c, s, e, v in self.outlines:
                out.write('\t'.join([pre+c, str(s*self.binsize), str(e*self.binsize), str(v)])+'\n')

    
    def pieces(self, arr, minseq=10, scale='log'):

        idx = np.where(arr!=0)[0]
        tmp = np.split(idx, np.where(np.diff(idx)!=1)[0]+1)
        seqs = []
        for i in tmp:
            if i.size >= minseq:
                if scale == 'log':
                    seqs.append(np.log2(arr[i]))
                else:
                    seqs.append(arr[i])
        
        return seqs

    def _segment(self, arr, components=2):

        nonzero = arr[arr > 0]
        idx = self.hampel_filter(np.log2(nonzero))
        filtered = nonzero[idx]

        log_gmm = self.get_states(np.log2(filtered))
        log_means, log_probs = log_gmm.means_.ravel(), log_gmm.weights_
        ln_gmm = self.get_states(filtered) # to improve the sensitivity
        ln_means, ln_probs = ln_gmm.means_.ravel(), ln_gmm.weights_
        if (len(log_means) == 1):
            means, probs = ln_means, ln_probs
            scale = 'linear'
        else:
            means, probs = log_means, log_probs
            scale = 'log'

        logger.info('Estimated HMM state number: {0} ({1} scale)'.format(len(means), scale))
        model = HiddenMarkovModel()
        # GMM emissions
        dists = []
        for m in means:
            tmp = []
            for i in range(components):
                e = m + (-1)**i * ((i+1)//2) * 0.5
                s = 0.5
                tmp.append(NormalDistribution(e, s))
            mixture = State(GeneralMixtureModel(tmp), name=str(m))
            dists.append(mixture)
        model.add_states(*tuple(dists))
        # transition matrix
        for i in range(len(means)):
            for j in range(len(means)):
                if i==j:
                    model.add_transition(dists[i], dists[j], 0.8)
                else:
                    model.add_transition(dists[i], dists[j], 0.2/(len(means)-1))
        
        # starts and ends
        for i in range(len(means)):
            model.add_transition(model.start, dists[i], probs[i])
        
        model.bake()

        # training sequences
        tmp = np.zeros(nonzero.size)
        tmp[idx] = filtered
        newarr = np.zeros(arr.size)
        newarr[arr > 0] = tmp

        if len(means) > 1:
            model.fit(self.pieces(newarr, scale=scale), algorithm='baum-welch', n_jobs=self.n_jobs,
                    max_iterations=5000, stop_threshold=2e-4)
            
            queue = newarr[newarr > 0]
            
            if scale=='log':
                seq = np.r_[[s.name for i, s in model.viterbi(np.log2(queue))[1][1:]]]
            else:
                seq = np.r_[[s.name for i, s in model.viterbi(queue)[1][1:]]]
            seg = self.assign_cnv(queue, seq)
            
            predicted = np.zeros(newarr.size)
            predicted[newarr > 0] = seg
            seg = self.call_intervals(predicted)
        else:
            seg = [(0, newarr.size)]
        
        return newarr, seg, scale
    
    def assign_cnv(self, cnv_profile, hmmseq):

        seg = np.zeros(cnv_profile.size)
        for i in np.unique(hmmseq):
            idx = np.where(hmmseq == i)[0]
            tmp = np.split(idx, np.where(np.diff(idx)!=1)[0]+1)
            for t in tmp:
                local_mean = np.median(cnv_profile[t])
                seg[t] = local_mean
        
        return seg
    
    def call_intervals(self, cnvs):

        seg = []
        for i in np.unique(cnvs):
            idx = np.where(cnvs == i)[0]
            tmp = np.split(idx, np.where(np.diff(idx)!=1)[0]+1)
            for t in tmp:
                seg.append((t[0], t[-1]+1))
        
        seg.sort()

        return seg
    
    def get_states(self, arr, maxdist=15):
        """
        Chromosome level estimation.
        """
        X = arr[:, np.newaxis]

        n_components_range = range(1, maxdist+1)
        _trace_gmm = {}
        bic_arr = []
        for n_components in n_components_range:
            gmm = GaussianMixture(n_components=n_components,
                                covariance_type='diag')
            gmm.fit(X)
            bic = gmm.bic(X)
            bic_arr.append(bic)
            _trace_gmm[n_components] = gmm
        
        bic_arr = np.r_[bic_arr]
        best_idx = bic_arr.argmin()
        #loss_rate = np.abs((bic_arr - bic_arr.min()) / bic_arr.min())
        #best_idx = np.where(loss_rate < 0.01)[0][-1]
        
        gmm = _trace_gmm[best_idx+1]
        
        return gmm

    
    def hampel_filter(self, arr, ws = 10, n_sigmas = 3):
    
        n = len(arr)
        k = 1.4826 # scale factor for Gaussian distribution
        
        indices = []
        
        for i in range(n):
            if i < ws:
                si = 0
            else:
                si = i - ws
            if i + ws > n:
                ei = n
            else:
                ei = i + ws
            sub = arr[si:ei]
            x0 = np.median(sub)
            S0 = k * np.median(np.abs(sub - x0))
            
            if np.abs(arr[i] - x0) < n_sigmas * S0:
                indices.append(i)
        
        return indices
        


#cython: language_level=3
#cython: boundscheck=False
#cython: cdivision=True

import logging
from tadlib.domaincaller.chromLev import Chrom as minChrom
from scipy.sparse import csr_matrix
import numpy as np

log = logging.getLogger(__name__)

    
class TADcaller(minChrom):

    def __init__(self, hicdata, res, model, window_size=2000000):

        hicdata = csr_matrix(hicdata, shape=hicdata.shape)
        minChrom.__init__(self, 'chrN', res, hicdata)
        ws = min(window_size // res, hicdata.shape[0]-5)
        self._dw = ws
        self.windows = np.ones(hicdata.shape[0], np.int32) * ws
        self.hmm = model
    
    def callDomains(self):

        self.calDI(self.windows, 0)
        self.splitChrom(self.DIs)
        minDomains = self.minCore(self.regionDIs)
        self.domains = self.getDomainList(minDomains)
    
    def loop_like(self):

        L = []
        for d in self.domains:
            L.append((d[0]//self.res, d[1]//self.res, 0))
        
        L.sort()

        return L




#cython: language_level=3
#cython: boundscheck=False
#cython: cdivision=True

import itertools, matplotlib, neoloop, os
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap, LogNorm
from matplotlib.gridspec import GridSpec
from neoloop.visualize.bed import Genes, plotGenes, Elements
from neoloop.visualize.bigwig import plotSignal, SigTrack
from neoloop.visualize.loops import Loops
from neoloop.assembly import complexSV
from neoloop.callers import Peakachu
from scipy import sparse

new_rc_params = {'text.usetex': False,
"svg.fonttype": 'none'
}

matplotlib.rcParams.update(new_rc_params)

class Triangle(Peakachu):

    track_colors = {
        'CNV': '#666666',
        'RNA+': '#E31A1C',
        'RNA-': '#E31A1C',
        'RNA': '#E31A1C',
        'H3K27ac': '#6A3D9A',
        'DNase': '#6A3D9A',
        '4C': '#33A02C'
    }

    chrom_colors = {'chr1': '#B15928',
            'chr2': '#6A3D9A',
            'chr3': '#CAB2D6',
            'chr4': '#FDBF6F',
            'chr5': '#A6CEE3',
            'chr6': '#B2DF8A',
            'chr7': '#1B9E77',
            'chr8': '#A6CEE3',
            'chr9': '#33A02C',
            'chr10':'#E6AB02',
            'chr11':'#E58606',
            'chr12':'#5D69B1',
            'chr13':'#FF7F00',
            'chr14':'#52BCA3',
            'chr15':'#99C945',
            'chr16':'#CC61B0',
            'chr17':'#88CCEE',
            'chr18':'#1D6996',
            'chr19':'#117733',
            'chr20':'#661100',
            'chr21':'#882255',
            'chr22':'#1F78B4',
            'chrX':'#666666',
            'chrY':'#5F4690'
            }

    def __init__(self, clr, assembly, span=None, protocol='insitu', slopes={},
        n_rows=3, track_partition=[4,0.1,0.3], space=0.04, correct='sweight', figsize=(7, 4)):
        """
        Recommended parameter settings.

        Set1: (Contact map + genes + chromosome bar)
            figsize=(7, 4)
            track_partition=[5, 0.2, 0.5]
            n_rows=3
            space=0.02
        
        Set2: (Contact map + genes + 1 signal track + chromosome bar)
            figsize=(7, 4.6)
            track_partition=[5, 0.2, 0.8, 0.5]
            n_rows=4
            genes: fontsize=7
            signal: data_range_size=7
            chromosome bar: name_size=7
        
        Set2_2: (Contact map + genes + cREs + chromosome bar)
            figsize=(7, 4.4)
            track_partition=[5, 0.3, 0.3, 0.5]
            n_rows=4
            genes: fontsize=7
            signal: data_range_size=7
            chromosome bar: name_size=7
        
        Set3: (Contact map + genes + 2 signal tracks + chromosome bar)
            figsize=(7, 5.2)
            track_partition=[5, 0.2, 0.8, 0.8, 0.5]
            n_rows=5
        
        Set4: (Contact map + genes + 3 signal tracks + chromosome bar)
            figsize=(7, 5.8)
            track_partition=[5, 0.2, 0.8, 0.8, 0.8, 0.5]
            n_rows=6
        
        Set5: (Contact map + genes + 4 signal tracks + chromosome bar)
            figsize=(7, 6.4)
            track_partition=[5, 0.2, 0.8, 0.8, 0.8, 0.8, 0.5]
            n_rows=7 
        
        Set6: (Contact map + genes + 5 signal tracks + chromosome bar)
            figsize=(7,7)
            track_partition=[5, 0.2, 0.8, 0.8, 0.8, 0.8, 0.8, 0.5]
            n_rows=8
        
        Set7: (Contact map + Arcs + 4C + genes + chromosome bar)
            figsize=(7, 5.2)
            track_partition=[5, 0.8, 0.8, 0.2, 0.5]
            n_rows=5

        """
        parse = assembly.split()
        self.assembly_ID = parse[0]
        if not span is None:
            assembly = complexSV(clr, '\t'.join(parse[1:]), col=correct, protocol=protocol,
                                 span=span, flexible=False, slopes=slopes)
        else:
            assembly = complexSV(clr, '\t'.join(parse[1:]), col=correct, protocol=protocol,
                                slopes=slopes)
        assembly.reorganize_matrix()
        assembly.correct_heterozygosity()
        self.assembly = assembly

        fig = plt.figure(figsize=figsize)
        self.fig = fig
        self.grid = GridSpec(n_rows, 1, figure=fig, left=0.1, right=0.9,
                    bottom=0.1, top=0.9, hspace=space, height_ratios=track_partition)
        self.track_count = 0
        self.w_h = figsize[0] / figsize[1]

        self.matrix = assembly.symm_hcm # should be a symmetric matrix
        self.bounds = assembly.bounds
        self.orients = assembly.orients
        self.Map = assembly.Map
        self.res = clr.binsize
        self.sig_data = {}
        self.sig_tracks = {}

        # define my colormap (traditional w --> r)
        self.cmap = LinearSegmentedColormap.from_list('interaction',
                ['#FFFFFF','#FFDFDF','#FF7575','#FF2626','#F70000'])
    
    def print_coordinate(self, pos):

        if pos % 1000000 == 0:
            return '{0}M'.format(pos//1000000)
        else:
            return '{0:.2f}M'.format(pos/1000000)
    
    def matrix_plot(self, colormap='traditional', vmin=None, vmax=None, log=False,
        cbr_width=0.02, cbr_height=0.1, cbr_fontsize=7, no_colorbar=False):

        h_ax = self.fig.add_subplot(self.grid[self.track_count])
        self.track_count += 1

        heatmap_pos = h_ax.get_position().bounds

        M = self.matrix
        n = M.shape[0]

        # Create the rotation matrix
        t = np.array([[1,0.5], [-1,0.5]])
        A = np.dot(np.array([(i[1],i[0]) for i in itertools.product(range(n,-1,-1),range(0,n+1,1))]),t)

        if colormap=='traditional':
            cmap = self.cmap
        else:
            cmap = colormap
        
        # Plot the Heatmap ...
        x = A[:,1].reshape(n+1, n+1)
        y = A[:,0].reshape(n+1, n+1)
        y[y<0] = -y[y<0]

        if vmax is None:
            vmax = np.percentile(M[M.nonzero()], 95)
        if vmin is None:
            vmin = M.min()
        
        if log:
            vmin = M[np.nonzero(M)].min()
            vmax = M.max()
            sc = h_ax.pcolormesh(x, y, np.flipud(M), cmap=cmap,
                        edgecolor='none', snap=True, linewidth=.001, norm=LogNorm(vmin, vmax))
        else:
            sc = h_ax.pcolormesh(x, y, np.flipud(M), vmin=vmin, vmax=vmax, cmap=cmap,
                        edgecolor='none', snap=True, linewidth=.001)
        
        h_ax.axis('off')
        self.heatmap_ax = h_ax
        self.hx = x
        self.hy = y
        
        # colorbar
        if not no_colorbar:
            c_ax = self.fig.add_axes([heatmap_pos[0]-0.02,
                                 (heatmap_pos[1]+0.9)/2,
                                 cbr_width,
                                 cbr_height])
            cbar = self.fig.colorbar(sc, cax=c_ax, ticks=[vmin, vmax], format='%.3g')
            c_ax.tick_params(labelsize=cbr_fontsize)
            self.cbar_ax = c_ax
        
    
    def plot_chromosome_bounds(self, line_color='k', linewidth=1.5, linestype=':'):

        n = self.matrix.shape[0]

        bounds = []
        for i in range(2, len(self.bounds), 2):
            bounds.append(self.bounds[i][0])
        
        for si in bounds:
            ei = n - 1
            x = [self.hx[:-1, :-1][n-1-si, si],
                 self.hx[:-1, :-1][n-1-si, ei]]
            y = [self.hy[:-1, :-1][n-1-si, si] - 1,
                 self.hy[:-1, :-1][n-1-si, ei] + 1]
            self.heatmap_ax.plot(x, y, color=line_color, linestyle=linestype,
                linewidth=linewidth)
        
        for ei in bounds:
            si = 0
            x = [self.hx[:-1, :-1][n-1-si, ei],
                 self.hx[:-1, :-1][n-1-ei, ei]]
            y = [self.hy[:-1, :-1][n-1-si, ei] + 1,
                 self.hy[:-1, :-1][n-1-ei, ei] - 1]
            self.heatmap_ax.plot(x, y, color=line_color, linestyle=linestype,
                linewidth=linewidth)
        
        self.heatmap_ax.set_xlim(self.hx.min(), self.hx.max())
        self.heatmap_ax.set_ylim(self.hy.min(), self.hy.max())
    
    def plot_neoTAD(self, ws=500000, color='#60636A'):

        from neoloop.tadtool.core import TADcaller
        import joblib

        hmm_folder = os.path.join(os.path.split(neoloop.__file__)[0], 'data')
        hmm = joblib.load(os.path.join(hmm_folder, 'HMM-model.pkl'))

        work = TADcaller(self.matrix, self.res, hmm, window_size=ws)
        work.callDomains()
        self.tads = work.loop_like()

        pairs = []
        for st, et, _ in self.tads:
            pairs.append((st+0.5, et+0.5))
            
        pairs.sort()

        tad_ax = self.fig.add_subplot(self.grid[self.track_count])
        self.track_count += 1
        tad_ax.set_xlim(self.hx.min(), self.hx.max())
        bottoms = np.r_[[i%2 for i in range(len(pairs))]]
        widths = np.r_[[d[1]-d[0] for d in pairs]]
        lefts = np.r_[[d[0] for d in pairs]]
        tad_ax.barh(bottoms, widths, 0.8, lefts, color=color,
                    edgecolor='none')
        tad_ax.set_ylim(-0.5, 2.1)
        tad_ax.set_axis_off()
        self.tad_ax = tad_ax
    
    def plot_loops(self, loop_fil, marker_size=50, face_color='#1F78B4', edgecolors='#1F78B4', 
        marker_type='o', marker_alpha=1, cluster=True, onlyneo=False, filter_by_res=True):

        self.M = sparse.csr_matrix(self.matrix)
        self.r = self.res

        loops = Loops(loop_fil)
        self.loops = loops.loops[self.assembly_ID]
        
        Donuts = {}
        Bool = np.zeros(self.matrix.shape, dtype=bool)
        for l1, l2, label in self.loops:
            # Lodate the loop pixel at given resolution
            xs, xe = l1[1:]
            ys, ye = l2[1:]
            if filter_by_res:
                if xe - xs != self.res:
                    continue
            s_l = range(xs//self.res-1, int(np.ceil(xe/float(self.res)))+1)
            e_l = range(ys//self.res-1, int(np.ceil(ye/float(self.res)))+1)
            si, ei = None, None
            for i in s_l:
                for j in e_l:
                    loci1 = (l1[0], i*self.res)
                    loci2 = (l2[0], j*self.res)
                    if (loci1 in self.Map) and (loci2 in self.Map):
                        st, et = self.Map[loci1], self.Map[loci2]
                        if st > et:
                            st, et = et, st
                        if label:
                            if self.assembly.chains[st] == self.assembly.chains[et]:
                                continue
                        else:
                            if self.assembly.chains[st] != self.assembly.chains[et]:
                                continue
                        if si is None:
                            si, ei = st, et
                        else:
                            if self.matrix[st,et] > self.matrix[si,ei]:
                                si, ei = st, et
            if not si is None:
                if onlyneo:
                    if label:
                        Donuts[(si, ei)] = self.matrix[si, ei]
                else:
                    Donuts[(si, ei)] = self.matrix[si, ei]
        
        if cluster:
            Donuts = self.local_clustering(Donuts, min_count=1, r=20000, chains=self.assembly.chains)
        
        for si, ei in Donuts:
            Bool[si, ei] = 1
        
        lx = self.hx[:-1,:-1][np.flipud(Bool)]
        ly = self.hy[:-1,:-1][np.flipud(Bool)] + 1
        self.Bool = Bool

        if lx.size > 0:
            self.heatmap_ax.scatter(lx, ly, s=marker_size, c=face_color, marker=marker_type,
                edgecolors=edgecolors, alpha=marker_alpha)
        
        self.heatmap_ax.set_xlim(self.hx.min(), self.hx.max())
        self.heatmap_ax.set_ylim(self.hy.min(), self.hy.max())
    
    def plot_DI(self, ws=2000000, pos_color='#FB9A99', neg_color='#A6CEE3', y_axis_offset=0.01,
        data_range_size=7):

        from tadlib.hitad.chromLev import Chrom
        from scipy.sparse import csr_matrix
        from neoloop.visualize.bigwig import plot_y_axis

        di_ax = self.fig.add_subplot(self.grid[self.track_count])
        self.track_count += 1
        
        hicdata = csr_matrix(self.matrix, shape=self.matrix.shape)
        tad_core = Chrom('chrN', self.res, hicdata, 'pseudo')
        tad_core._dw = min(tad_core._dw, hicdata.shape[0]-1)
        tad_core.windows = np.ones(hicdata.shape[0], dtype=int) * (ws // self.res)
        tad_core.calDI(tad_core.windows, 0)
        arr = tad_core.DIs

        pos_mask = arr >= 0
        neg_mask = arr < 0
        di_ax.fill_between(np.arange(arr.size), arr, where=pos_mask, color=pos_color,
                           edgecolor='none')
        di_ax.fill_between(np.arange(arr.size), arr, where=neg_mask, color=neg_color,
                           edgecolor='none')
        di_ax.set_xlim(self.hx.min(), self.hx.max())

        ymin, ymax = di_ax.get_ylim()
        ax_pos = di_ax.get_position().bounds
        y_ax = self.fig.add_axes([ax_pos[0]-y_axis_offset, ax_pos[1],
                                y_axis_offset, ax_pos[3]])
        plot_y_axis(y_ax, ymin, ymax, size=data_range_size)
        self.clear_frame(y_ax)
        self.clear_frame(di_ax)

        self.di_ax = di_ax
        self.DI_arr = arr


    def plot_arcs(self, h_ratio=0.3, arc_color='#386CB0', arc_alpha=1,
        lw=1.5, cutoff='bottom', species='human', release=97, gene_filter=[]):
        """
        Both plot_loops and plot_genes need to be called before this method.
        """
        from matplotlib.patches import Arc

        arc_ax = self.fig.add_subplot(self.grid[self.track_count])
        self.track_count += 1
        arc_ax.set_xlim(self.hx.min(), self.hx.max())

        xs, ys = np.where(self.Bool)
        if not len(gene_filter):
            x, y = xs, ys
        else:
            # filter loops by genes
            refgenes = Genes(self.bounds, self.orients, self.res, species=species, release=release,
                         filter_=gene_filter).genes
            pos = set()
            for g in gene_filter:
                for t in refgenes:
                    if t[3] == g:
                        for i in range(t[1]//self.res, t[2]//self.res+1):
                            pos.add(i)
            x = []; y = []
            for i, j in zip(xs, ys):
                if (i in pos) or (j in pos):
                    x.append(i+0.5)
                    y.append(j+0.5)
        
        miny = maxy = 0
        for i, j in zip(x, y):
            mid = (i + j) / 2
            a, b = j - i, (j - i) * h_ratio
            if cutoff=='bottom':
                ty = b / 2 + 0.5
                if ty > maxy:
                    maxy = ty
                arc_ax.add_patch(Arc((mid, 0), a, b, 
                             theta1=0.0, theta2=180.0, edgecolor=arc_color, lw=lw, alpha=arc_alpha))
            else:
                ty = -b / 2 - 0.5
                if ty < miny:
                    miny = ty
                arc_ax.add_patch(Arc((mid, 0), a, b, 
                             theta1=180.0, theta2=360.0, edgecolor=arc_color, lw=lw, alpha=arc_alpha))
        
        arc_ax.set_ylim(miny, maxy)
        arc_ax.set_axis_off()
        self.arc_ax = arc_ax
        self.arc_x = x
        self.arc_y = y
            

    def plot_chromosome_bar(self, coord_ypos=0.3, name_ypos=0, coord_size=3.5, name_size=7,
        width=3, headwidth=8, remove_coord=False, color_by_order=[]):

        chrom_ax = self.fig.add_subplot(self.grid[self.track_count])
        self.track_count += 1
        chrom_colors = self.chrom_colors

        chrom_ax.set_xlim(self.hx.min(), self.hx.max())
        chrom_ax.set_ylim(0, 1)
        # chromosome bar
        n = self.matrix.shape[0]
        for i in range(0, len(self.bounds), 2):
            s = self.bounds[i][0]
            e = self.bounds[i+1][0]
            si = self.hx[:-1, :-1][n-1-s, s]
            ei = self.hx[:-1, :-1][n-1-e, e]
            o = self.orients[i//2]
            c = 'chr'+self.bounds[i][1][0].lstrip('chr')
            bar_color = chrom_colors[c]
            if len(color_by_order):
                bar_color = color_by_order[i//2]
            if o=='+':
                chrom_ax.annotate('', xy=(ei, 0.7), xytext=(si, 0.7),
                        xycoords='data', arrowprops=dict(color=bar_color, shrink=0, width=width, headwidth=headwidth),
                    )
            else:
                chrom_ax.annotate('', xy=(si, 0.7), xytext=(ei, 0.7),
                        xycoords='data', arrowprops=dict(color=bar_color, shrink=0, width=width, headwidth=headwidth),
                    )
            
            if not remove_coord:
                chrom_ax.text(si, coord_ypos, self.print_coordinate(self.bounds[i][1][1]),
                        ha='left', va='top', fontsize=coord_size)
                chrom_ax.text(ei, coord_ypos, self.print_coordinate(self.bounds[i+1][1][1]),
                        ha='right', va='top', fontsize=coord_size)
        
        # concatenate chromosome labels
        chroms = []
        for b in self.bounds:
            if not len(chroms):
                chroms.append([b[1][0], b[0], b[0]])
            else:
                if b[1][0] == chroms[-1][0]:
                    chroms[-1][-1] = b[0]
                else:
                    chroms.append([b[1][0], b[0], b[0]])
        
        for c in chroms:
            si = self.hx[:-1, :-1][n-1-c[1], c[1]]
            ei = self.hx[:-1, :-1][n-1-c[2], c[2]]
            chrom_ax.text((si+ei)/2, name_ypos, 'chr'+c[0].lstrip('chr'),
                    ha='center', va='top', fontsize=name_size)
        
        chrom_ax.axis('off')
    
    def plot_genes(self, species='human', release=97, filter_=None,
        color='#999999', border_color='#999999', fontsize=7, labels='auto',
        style='flybase', global_max_row=False, label_aligns={}):

        genes = Genes(self.bounds, self.orients, self.res, species=species, release=release,
                filter_=filter_)
        _wk = plotGenes(genes.file_handler, color=color, fontsize=fontsize, labels=labels,
                style=style, global_max_row=global_max_row)
        
        ax = self.fig.add_subplot(self.grid[self.track_count])
        self.track_count += 1

        _wk.plot(ax, 'chrN', 0, self.matrix.shape[0]*self.res, label_aligns)

        ax.set_xlim(-self.res/2, self.matrix.shape[0]*self.res+self.res/2)
        ax.axis('off')

        self.gene_ax = ax
        self.genes = genes
    
    def plot_cREs(self, bedfil, color='#E41A1C', alpha=0.5):

        ax = self.fig.add_subplot(self.grid[self.track_count])
        self.track_count += 1

        eles = Elements(bedfil, self.bounds, self.orients, self.res)
        cREs = eles.cREs

        bottoms = np.zeros(len(cREs))
        widths = np.r_[[d[2]-d[1] for d in cREs]]
        lefts = np.r_[[d[1]+self.res/2 for d in cREs]]
        ax.barh(bottoms, widths, 0.8, lefts, color=color,
                edgecolor=color, alpha=alpha, linewidth=0)
        ax.set_ylim(-0.2, 1)
        ax.set_axis_off()
        ax.set_xlim(0, self.matrix.shape[0]*self.res+self.res)

        self.cRE_ax = ax
    
    def plot_motif(self, bedfil, plus_color='#444444', minus_color='#999999', ypos=0.5,
        marker_size=10, subset='+'):

        ax = self.fig.add_subplot(self.grid[self.track_count])
        self.track_count += 1

        eles = Elements(bedfil, self.bounds, self.orients, self.res)
        motifs = eles.cREs
        for m in motifs:
            x = (m[1] + m[2]) // 2
            strand = m[3]
            if strand != subset:
                continue
            if strand == '+':
                ax.scatter(x, ypos, s=marker_size, c=plus_color, marker='>')
            else:
                ax.scatter(x, ypos, s=marker_size, c=minus_color, marker='<')

        ax.set_ylim(0, 1)
        ax.set_axis_off()
        ax.set_xlim(0, self.matrix.shape[0]*self.res+self.res)
        self.motif_ax = ax
        
    
    def clear_frame(self, ax):

        for spine in ax.spines:
            ax.spines[spine].set_visible(False)
        
        ax.tick_params(axis='both', bottom=False, top=False, left=False, right=False,
            labelbottom=False, labeltop=False, labelleft=False, labelright=False)

    
    def plot_signal(self, track_name, bw_fil, factor=1, ax=None, color='auto', show_data_range=True,
        data_range_style='y-axis', data_range_size=7, max_value='auto', min_value='auto',
        y_axis_offset=0.01, show_label=True, label_size=7, label_pad=2, nBins=500,
        style_params={'type':'fill'}):
        '''
        Choices for data_range_style: ['y-axis', 'text'].
        '''
        sigs = SigTrack(bw_fil, self.bounds, self.orients, res=self.res, nBins=nBins,
                        multiply=factor)
        arr = sigs.stats('chrN', 0, self.matrix.shape[0]*self.res, 100)

        if color=='auto':
            if track_name in self.track_colors:
                color = self.track_colors[track_name]
            else:
                color = '#dfccde'

        if max_value=='auto':
            max_value = np.nanmax(arr)
        
        if min_value=='auto':
            min_value = 0

        _wk = plotSignal(sigs, color=color, show_data_range='no',
                        max_value=max_value, min_value=min_value,
                        number_of_bins=self.matrix.shape[0] * sigs.factor)
        
        if ax is None:
            ax = self.fig.add_subplot(self.grid[self.track_count])
            self.track_count += 1
        else:
            ax = ax

        _wk.properties.update(style_params)
        _wk.plot(ax, 'chrN', 0, self.matrix.shape[0] * self.res)
        self.clear_frame(ax)

        if show_data_range:
            if data_range_style=='y-axis':
                ax_pos = ax.get_position().bounds
                y_ax = self.fig.add_axes([ax_pos[0]-y_axis_offset, ax_pos[1],
                                        y_axis_offset, ax_pos[3]])
                _wk.plot_y_axis(y_ax, size=data_range_size)
                self.clear_frame(y_ax)
            else:
                _wk.plot_range_text(min_value, max_value, data_range_size)
        
        if show_label:
            if show_data_range and (data_range_style=='y-axis'):
                y_ax.set_ylabel(track_name, rotation=0, va='center', ha='right',
                    fontsize=label_size, labelpad=label_pad)
            else:
                ax.set_ylabel(track_name, rotation=0, va='center', ha='right', 
                    fontsize=label_size, labelpad=label_pad)
        
        ax.set_xlim(-self.res/2, self.matrix.shape[0]*self.res+self.res/2)

        self.sig_data[track_name] = sigs
        self.sig_tracks[track_name] = ax


    def outfig(self, outfile, dpi=200, bbox_inches='tight'):

        self.fig.savefig(outfile, dpi=dpi, bbox_inches=bbox_inches)
    
    def show(self):

        self.fig.show()

def plot_exp(exp, res, outfig, dpi=200):

    x = []
    y = []
    for i in sorted(exp):
        x.append(i*res)
        y.append(exp[i])
    
    fig = plt.figure(figsize=(5, 5))
    ax = fig.add_subplot(111)

    l1, = ax.plot(x, y)
    ax.set_xlabel('Genomic distance (bp)')
    ax.set_ylabel('Average contact frequency')

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_linewidth(1.5)
    ax.spines['left'].set_linewidth(1.5)

    ax.xaxis.set_tick_params(width=1.5)
    ax.yaxis.set_tick_params(width=1.5)

    ax.ticklabel_format(scilimits=(-3,3))

    plt.savefig(outfig, dpi=dpi, bbox_inches='tight')
    plt.close()



#cython: language_level=3
#cython: boundscheck=False
#cython: cdivision=True

from neoloop.callers import Fusion
from neoloop.cnv.loadcnv import binCNV
from neoloop.util import find_chrom_pre
import numpy as np
from joblib import Parallel, delayed
import os

def outside_trans(arg, **kwarg):

    return Train.trans_sum(*arg, **kwarg)


def outside_cis(arg, **kwarg):

    return Train.cis_vec(*arg, **kwarg)


class Train(binCNV, Fusion):

    def __init__(self, clr, cnv_file, n_jobs = 1):
        
        self.clr = clr
        self.res = self.clr.binsize
        self.n_jobs = n_jobs

        binCNV.__init__(self, cnv_file, self.res)

        self.queue = {}
        self.pre = find_chrom_pre(clr.chromnames)
        for c in self.bin_cnv:
            tmp = self.seg(c)
            for v in tmp:
                if not v in self.queue:
                    self.queue[v] = []
                self.queue[v].extend(tmp[v])
    
    def expected_by_cnv(self):

        # Parallel trans stats computation
        trans = {}
        for ci in self.queue:
            for cj in self.queue:
                if ci > cj:
                    continue
                key = (ci, cj)
                params = set()
                for r1 in self.queue[ci]:
                    for r2 in self.queue[cj]:
                        rr1, rr2 = r1, r2
                        if rr1 > rr2:
                            rr1, rr2 = rr2, rr1
                        if rr1[0] == rr2[0]:
                            continue
                        params.add((self, rr1, rr2))
                if len(params):
                    results = Parallel(n_jobs=self.n_jobs)(delayed(outside_trans)(i) for i in params)
                    results = np.r_[results]
                    counts = results[:,0].sum()
                    num = results[:,1].sum()
                    if num > 0:
                        trans[key] = [counts / num, num]
                    else:
                        trans[key] = [0, 0]
                else:
                    trans[key] = [0, 0]

        self.trans = trans
        
        # Parallel cis stats computation
        cis = {}
        for ci in self.queue:
            for cj in self.queue:
                if ci > cj:
                    continue
                key = (ci, cj)
                params = set()
                for r1 in self.queue[ci]:
                    for r2 in self.queue[cj]:
                        rr1, rr2 = r1, r2
                        if rr1 > rr2:
                            rr1, rr2 = rr2, rr1
                        if rr1[0] != rr2[0]:
                            continue
                        params.add((self, rr1, rr2))
                tmp = {}
                if len(params):
                    results = Parallel(n_jobs=self.n_jobs)(delayed(outside_cis)(i) for i in params)
                    for D in results:
                        for i in D:
                            if not i in tmp:
                                tmp[i] = [0, 0]
                            tmp[i][0] += D[i][0]
                            tmp[i][1] += D[i][1]
                cis[key] = {}
                for i in tmp:
                    if tmp[i][1] >= 3:
                        cis[key][i] = [tmp[i][0] / tmp[i][1], tmp[i][1]]
        
        self.cis = cis
    
    def trans_factor(self, ref = 3):

        tmp = {}
        ref_v = self.trans[(ref, ref)][0]
        for i in self.trans:
            f1 = self.trans[i][0] / ref_v
            tmp[i] = f1
        
        return tmp
    
    def cis_factor(self, ref = 3):

        slopes = {}
        intervals = [(0, 20000000//self.res),
                     (20000000//self.res, 500000000//self.res)]
        
        k = (ref, ref)
        ref_e = self.cis[k]
        for i in sorted(self.cis):
            slopes[i] = [[1, 1]]
            if i == k:
                slopes[i] = [[1, 1]]
            else:
                cur = self.cis[i]
                pool = []
                for minx, maxx in intervals:
                    exp2 = {t:ref_e[t][0] for t in ref_e if (t >= minx) and (t <= maxx)}
                    exp1 = {t:cur[t][0] for t in cur if (t >= minx) and (t <= maxx)}
                    cr, cs, w = self.linear_regression(exp1, exp2)
                    pool.append([cs, cr])
                slopes[i] = pool

        return slopes
    
    def seg(self, chrom, minsize=2):

        arr = self.bin_cnv[chrom]
        chrom = self.pre + chrom
        chromLen = self.clr.chromsizes[chrom]
        values = np.unique(arr)
        D = {} # cnv-centered segments
        for v in values:
            tmp = np.where(arr==v)[0]
            pieces = np.split(tmp, np.where(np.diff(tmp)!=1)[0]+1)
            filtered = [p for p in pieces if len(p) >= minsize]
            intervals = []
            for p in filtered:
                intervals.append((chrom, p[0]*self.res, min((p[-1]+1)*self.res, chromLen)))
            
            if len(intervals):
                key = float(v)
                D[key] = intervals
        
        return D
    
    def trans_sum(self, r1, r2):

        M = self.clr.matrix(balance=False, sparse=True).fetch(r1, r2).tocsr()
        M2 = self.clr.matrix(balance=True, sparse=False).fetch(r1, r2)
        x, y = np.where(np.isfinite(M2))

        if x.size > 0:
            total = M[x, y].sum()
            num = x.size
            return [total, num]
        else:
            return [0, 0]
        
    
    def cis_vec(self, r1, r2):
        '''
        r1 and r2 must be on the same chromosome, and r2 > r1
        '''
        M = self.clr.matrix(balance=False, sparse=True).fetch(r1, r2).tocsr()
        M2 = self.clr.matrix(balance=True, sparse=False).fetch(r1, r2)
        x, y = np.where(np.isfinite(M2))
        D = {}
        if x.size > 0:
            gx = x + r1[1]//self.res
            gy = y + r2[1]//self.res

            dis = gy - gx
            maxi = dis.max()
            mini = max(1, dis.min())
            for i in range(mini, maxi+1):
                mask = dis == i
                xi, yi = x[mask], y[mask]
                if xi.size > 0:
                    D[i] = [M[xi, yi].sum(), xi.size]
        
        return D


class SimCNV(Train):

    def __init__(self, clr, cnv_file, trans_factor={}, cis_factor={}, out_folder='SimCNV'):
        
        Train.__init__(self, clr, cnv_file)

        self.trans = trans_factor
        self.cis = cis_factor
        self.out_folder = out_folder
        if not os.path.exists(out_folder):
            os.mkdir(out_folder)
        self.cnv_by_chrom()
    
    def cnv_by_chrom(self):

        pairs = {}
        for i, c1 in enumerate(self.clr.chromnames):
            for j, c2 in enumerate(self.clr.chromnames):
                if i > j:
                    continue
                tmp = set()
                for k1 in self.queue:
                    for k2 in self.queue:
                        if k1 > k2:
                            continue
                        key = (k1, k2)
                        for r1 in self.queue[k1]:
                            for r2 in self.queue[k2]:
                                if (r1[0]==c1) and (r2[0]==c2):
                                    tmp.add((r1, r2, key))
                                elif (r2[0]==c1) and (r1[0]==c2):
                                    tmp.add((r2, r1, key))

                pairs[(c1, c2)] = tmp
        
        self.pairs = pairs
                        

    def correct_trans(self, chrom1, chrom2, ref=3):

        M = self.clr.matrix(balance=False).fetch(chrom1, chrom2)
        pairs = self.pairs[(chrom1, chrom2)]
        for r1, r2, k in pairs:
            if k==(ref, ref):
                continue
            s1, e1 = r1[1]//self.res, r1[2]//self.res
            s2, e2 = r2[1]//self.res, r2[2]//self.res
            f = self.trans[k]
            M[s1:e1, s2:e2] = M[s1:e1, s2:e2] * f
        
        self.out_matrix(M, chrom1, chrom2)
    
    def correct_cis(self, chrom, ref=3):

        M = self.clr.matrix(balance=False).fetch(chrom)
        pairs = self.pairs[(chrom, chrom)]

        intervals = [(0, 20000000//self.res),
                     (20000000//self.res, 500000000//self.res)]

        for r1, r2, k in pairs:
            if r1 > r2:
                continue
            s1, e1 = r1[1]//self.res, r1[2]//self.res
            s2, e2 = r2[1]//self.res, r2[2]//self.res
            fs = self.cis[k]
            if len(fs)!=len(intervals):
                continue

            f1, rsquare1 = fs[0]
            f2, rsquare2 = fs[1]
            if f1 > 0 and rsquare1 > 0.5:
                f = f1
            elif f2 > 0 and rsquare2 > 0.5:
                f = f2
            else:
                f = 1
            
            M[s1:e1, s2:e2] = M[s1:e1, s2:e2] * f
                    

        self.out_matrix(M, chrom, chrom)


    def out_matrix(self, M, chrom1, chrom2):

        outfil = os.path.join(self.out_folder, '{0}_{1}.txt'.format(chrom1.lstrip('chr'), chrom2.lstrip('chr')))
        if chrom1 == chrom2:
            M = np.triu(M)
        x, y = M.nonzero()
        values = M[x, y]
        with open(outfil, 'w') as out:
            for i, j, v in zip(x, y, values):
                line = '{0}\t{1}\t{2}\n'.format(i, j, v)
                out.write(line)
            
    


#cython: language_level=3
#cython: boundscheck=False
#cython: cdivision=True

import numpy as np
from scipy import optimize, sparse

def poisson_exp(X, counts, alpha, bias=None,
                beta=None, use_empty_entries=True):
    
    if bias is None:
        bias = np.ones((counts.shape[0], 1))

    ll = _poisson_exp_sparse(X, counts, alpha, bias=bias, beta=beta,
                            use_empty_entries=use_empty_entries)

    return ll

def _poisson_exp_sparse(X, counts, alpha, bias,
                        beta=None, use_empty_entries=False):
    m, n = X.shape

    d = np.sqrt(((X[counts.row] - X[counts.col])**2).sum(axis=1))
    if use_empty_entries:
        raise NotImplementedError

    bias = bias.flatten()
    if beta is None:
        beta = counts.sum() / (
            (d ** alpha) * bias[counts.row] * bias[counts.col]).sum()

    g = beta * d ** alpha * \
        bias[counts.row] * bias[counts.col]
    ll = (counts.data * np.log(beta) + alpha * counts.data * np.log(d) +
          counts.data * np.log(bias[counts.row] * bias[counts.col]))
    ll -= g

    if np.isnan(ll.sum()):
        raise ValueError("Objective function is Not a Number")

    return -ll.sum()

def gradient_poisson_exp(X, counts, alpha, bias=None,
                         beta=None, use_empty_entries=True):

    if bias is None:
        bias = np.ones((counts.shape[0], 1))

    return _gradient_poisson_exp_sparse(
            X, counts, alpha, bias, beta,
            use_empty_entries=use_empty_entries)

def _gradient_poisson_exp_sparse(X, counts, alpha, bias, beta,
                                 use_empty_entries=True):
    m, n = X.shape
    bias = bias.flatten()

    if use_empty_entries:
        raise NotImplementedError

    d = np.sqrt(((X[counts.row] - X[counts.col])**2).sum(axis=1))

    beta = counts.sum() / (
        (d ** alpha) * bias[counts.row] * bias[counts.col]).sum()

    grad_alpha = - beta * (bias[counts.row] * bias[counts.col] * d ** alpha *
                           np.log(d)).sum() \
        + (counts.data *
           np.log(d)).sum()

    return - np.array([grad_alpha])


def eval_f(x, user_data=None):

    m, n, counts, X, bias, use_empty_entries = user_data
    X = X.reshape((m, n))
    tmp = poisson_exp(X, counts, x[0], bias=bias,
                      use_empty_entries=use_empty_entries)
    return tmp


def eval_grad_f(x, user_data=None):
    """
    Evaluate the gradient of the function in alpha
    """

    m, n, counts, X, bias, use_empty_entries = user_data
    X = X.reshape((m, n))
    tmp = gradient_poisson_exp(X, counts, x[0], bias=bias, beta=None,
                               use_empty_entries=use_empty_entries)

    return tmp

def _estimate_beta(counts, X, alpha=-3, bias=None):

    if bias is None:
        bias = np.ones((counts.shape[0], 1))
    
    dis = np.sqrt(((X[counts.row] - X[counts.col])**2).sum(axis=1))
    bias = bias.flatten()
    beta = counts.sum() / (
            (dis ** alpha) * bias[counts.row] * bias[counts.col]).sum()
    
    return beta


def estimate_alpha_beta(counts, X, bias=None, ini=None,
                        use_empty_entries=False, random_state=None):

    m, n = X.shape
    bounds = np.array(
        [[-100, 1e-2]])

    if ini is None:
        ini = - random_state.randint(1, 100, size=(2, )) + \
            random_state.rand(1)

    data = (m, n, counts, X, bias,
            use_empty_entries)

    results = optimize.fmin_l_bfgs_b(
        eval_f,
        ini[0],
        eval_grad_f,
        (data, ),
        bounds=bounds,
        iprint=1,
        maxiter=1000,
        )

    beta = _estimate_beta(counts, X, alpha=results[0], bias=bias)

    return results[0], beta


from coolbox.core.track import Bed
from coolbox.plots.track import PlotBed
from coolbox.utilities import ReadBed, Interval, IntervalTree
from pyensembl import EnsemblRelease
import io, logging, bisect

log = logging.getLogger(__name__)

def get_blocks(bounds, orients, res):

    blocks = []
    for i in range(0, len(bounds), 2):
        b1 = bounds[i][1]
        b2 = bounds[i+1][1]
        if b1[1] > b2[1]:
            b1, b2 = b2, b1
        o = orients[i//2]
        s = bounds[i][0] * res
        blocks.append((s, b1[0].lstrip('chr'), b1[1], b2[1], o))
    
    return blocks


class Elements(object):
    '''
    Map cis regulatory elements to a new assembly.
    '''
    def __init__(self, bedfil, bounds, orients, res=10000, default_chrom='chrN'):

        self.blocks = get_blocks(bounds, orients, res)
        self.chrom_name = default_chrom
        self.elements = self.parse_bed(bedfil)
        self.map_elements()
    
    def map_elements(self):

        pool = set()
        for p, c, s, e, o in self.blocks:
            eles = self.extract_by_region('chr'+c, int(s), int(e))
            block_bounds = [p, p+e-s]
            for g in eles:
                if o == '-':
                    start = p + e - g[1]
                    end = p + e - g[2]
                else:
                    start = p + g[1] - s
                    end = p + g[2] - s

                if start > end:
                    start, end = end, start
                
                # truncate genes if it's beyond the original chromosome bounds
                if start < block_bounds[0]:
                    start = block_bounds[0]
                if end > block_bounds[1]:
                    end = block_bounds[1]
                
                if end <= start:
                    continue
                
                if len(g) > 3:
                    extra = list(g[3:])
                    if extra[0] == '+':
                        if o == '-':
                            extra[0] = '-'
                    elif extra[0] == '-':
                        if o == '-':
                            extra[0] = '+'
                    extra = tuple(extra)
                else:
                    extra = tuple()

                record = (self.chrom_name, start, end) + extra
                pool.add(record)
        
        self.cREs = sorted(pool)

    
    def parse_bed(self, bedfil):

        D =  {}
        with open(bedfil, 'r') as source:
            for line in source:
                p = line.rstrip().split()
                if not p[0] in D:
                    D[p[0]] = []
                D[p[0]].append([int(p[1]), int(p[2])] + p[3:])
        
        for c in D:
            D[c].sort()
        
        return D
    
    def extract_by_region(self, chrom, start, end):

        cache = set()
        List = self.elements[chrom]
        idx = max(0, bisect.bisect(List, [start, end])-1)
        for q in List[idx:]:
            if q[1] <= start:
                continue
            if q[0] >= end:
                break
            
            cache.add((chrom,) + tuple(q))
        
        return cache

class Genes(object):
    """
    Map reference genes onto a new assembly. 
    """
    def __init__(self, bounds, orients, res=10000, default_chrom='chrN',
        species='human', release=97, filter_=None):
        
        ref = EnsemblRelease(release, species=species)
        ref.download()
        ref.index()
        self.ref = ref
        if not filter_ is None:
            # a pre-defined gene list, such as cancer-related genes
            self.filter_list = set([g.upper() for g in filter_])
        else:
            self.filter_list = None

        self.blocks = get_blocks(bounds, orients, res)

        self.chrom_name = default_chrom

        self.map_genes()
        self.file_handler = io.StringIO('\n'.join(['\t'.join(list(map(str, g))) for g in self.genes]))
    
    def _check_overlap(self, i1, i2):

        if i1[0] != i2[0]:
            return 0, 0
        
        t1 = sorted(i1[1:])
        t2 = sorted(i2[1:])

        if (t1[1]<=t2[0]) or (t2[1]<=t1[0]):
            return 0, 0
        
        mi = t1 + t2
        mi.sort()
        OR = (mi[2] - mi[1]) / (mi[3] - mi[0]) # intersect / union

        return OR, mi[2]-mi[1]
    
    def map_genes(self):

        pool = []
        for p, c, s, e, o in self.blocks:
            genes = self.ref.genes_at_locus(c, int(s), int(e))
            block_bounds = [p, p+e-s]
            for g in genes:
                name = g.gene_name
                if not self.filter_list is None:
                    if not name.upper() in self.filter_list:
                        continue

                if not g.strand in ['+', '-']:
                    continue
                # gene body
                if o == '-':
                    start = p + e - g.start
                    end = p + e - g.end
                else:
                    start = p + g.start - s
                    end = p + g.end - s

                if start > end:
                    start, end = end, start
                
                # truncate genes if it's beyond the original chromosome bounds
                if start < block_bounds[0]:
                    start = block_bounds[0]
                if end > block_bounds[1]:
                    end = block_bounds[1]
                
                if end <= start:
                    continue

                if o == '+':
                    strand = g.strand
                else:
                    if g.strand == '+':
                        strand = '-'
                    else:
                        strand = '+'
                
                if len(g.exons):
                    # collect exons
                    exons = []
                    for en in g.exons:
                        size = en.end - en.start
                        if o == '-':
                            e_st = p + e - en.start
                            e_ed = p + e - en.end
                        else:
                            e_st = p + en.start - s
                            e_ed = p + en.end - s
                        if e_st > e_ed:
                            e_st, e_ed = e_ed, e_st
                        
                        # truncate exons
                        if e_st < start:
                            e_st = start
                        if e_ed > end:
                            e_ed = end
                        
                        if e_ed <= e_st:
                            continue

                        if e_st - start > 0:
                            exons.append((e_st-start, size)) # start relative to gene start
                    exons.sort()
                    exons = self._remove_redundant_exons(exons)
                    starts = ','.join([str(en[0]) for en in exons])+','
                    sizes = ','.join([str(en[1]) for en in exons])+','
                    exon_num = len(exons)
                else:
                    exon_num = 1
                    starts = '{0},'.format(0)
                    sizes = '{0},'.format(end-start)

                record = [self.chrom_name, start, end, name, 0, strand, start, end,
                        0, exon_num, sizes, starts]
                pool.append(record)
        
        self.genes = self._remove_redundant_genes(pool)
    
    def _remove_redundant_genes(self, genes):

        dedup = {g[3]:[0,0,0] for g in genes}
        for g in genes:
            Len = g[2] - g[1]
            cL = dedup[g[3]][2] - dedup[g[3]][1]
            if Len > cL:
                dedup[g[3]] = g
        
        genes = sorted([dedup[g] for g in dedup])

        return genes
    
    def _remove_redundant_exons(self, exons):

        pre_tree = {}
        for e in exons:
            if not e[0] in pre_tree:
                pre_tree[e[0]] = []
            pre_tree[e[0]].append(e[1])
        
        pre_exon = {e:max(pre_tree[e]) for e in pre_tree}

        nondup = []
        cur = 0
        for e in sorted(pre_exon):
            if e < cur:
                continue
            nondup.append((e, pre_exon[e]))
            cur = e + pre_exon[e]
        
        return nondup


class plotGenes(PlotBed):

    DEFAULT_COLOR = "#1f78b4"

    def __init__(self, file_, **kwargs):

        # here, file_ is an opened file or io.StringIO
        properties_dict = {
            'file_': file_,
            'file': 'inMemory.bed',
            'height': Bed.DEFAULT_HEIGHT,
            'color': 'bed_rgb',
            'border_color': 'black',
            'fontsize': Bed.DEFAULT_FONTSIZE,
            'title': '',
            'labels': 'auto',
            'style': 'flybase',
            'display': 'stacked',
            'global_max_row': False,
        }

        properties_dict.update(kwargs)

        labels = properties_dict.get('labels')
        if labels == 'auto':
            properties_dict['labels'] = 'auto'
        elif labels is True:
            properties_dict['labels'] = 'on'
        else:
            properties_dict['labels'] = 'off'
        
        self.properties = properties_dict
        
        self.bed_type = None  # once the bed file is read, this is bed3, bed6 or bed12
        self.len_w = None  # this is the length of the letter 'w' given the font size
        self.interval_tree = {}  # interval tree of the bed regions

        from matplotlib import font_manager

        self.fp = font_manager.FontProperties(size=self.properties['fontsize'])

        if 'interval_height' not in self.properties:
            self.properties['interval_height'] = 100

        if self.properties['labels'] != 'on':
            self.is_draw_labels = False
        else:
            self.is_draw_labels = True

        self.colormap = None

        # to set the distance between rows
        self.row_scale = self.properties['interval_height'] * 2.3

        self.interval_tree, min_score, max_score = self._process_bed()
    

    def _process_bed(self):

        bed_file_h = ReadBed(self.properties['file_'])
        self.bed_type = bed_file_h.file_type

        if 'color' in self.properties and self.properties['color'] == 'bed_rgb' and \
                self.bed_type not in ['bed12', 'bed9']:
            log.warning("*WARNING* Color set to 'bed_rgb', but bed file does not have the rgb field. The color has "
                        "been set to {}".format(PlotBed.DEFAULT_COLOR))
            self.properties['color'] = PlotBed.DEFAULT_COLOR

        valid_intervals = 0
        interval_tree = {}

        max_score = float('-inf')
        min_score = float('inf')
        for bed in bed_file_h:
            if bed.score < min_score:
                min_score = bed.score
            if bed.score > max_score:
                max_score = bed.score

            if bed.chromosome not in interval_tree:
                interval_tree[bed.chromosome] = IntervalTree()

            interval_tree[bed.chromosome].add(Interval(bed.start, bed.end, bed))
            valid_intervals += 1

        if valid_intervals == 0:
            log.warning("No valid intervals were found in file {}".format(self.properties['file_name']))

        return interval_tree, min_score, max_score
    
    def _get_length_w(self, fig_width, region_start, region_end):
        '''
        to improve the visualization of the genes it is good to have an estimation of the label
        length. In the following code I try to get the length of a 'W' in base pairs.
        '''
        if self.is_draw_labels:
            # from http://scipy-cookbook.readthedocs.org/items/Matplotlib_LaTeX_Examples.html
            inches_per_pt = 1.0 / 72.27
            font_in_inches = self.properties['fontsize'] * inches_per_pt
            region_len = region_end - region_start
            bp_per_inch = region_len / fig_width
            font_in_bp = font_in_inches * bp_per_inch
            self.len_w = font_in_bp
            log.debug("len of w set to: {} bp".format(self.len_w))
        else:
            self.len_w = 1

        return self.len_w
    
    def _get_y_pos(self, free_row):
        """
        The y_pos is set such that regions to be plotted do not overlap (stacked). To override this
        the properties['collapsed'] needs to be set.
        The algorithm uses a interval tree (self.region_interval) to check the overlaps
        and a sort of coverage vector 'rows used' to identify the row in which to plot
        Return
        ------
        ypos : int
            y position.
        """
        # if the domain directive is given, ypos simply oscilates between 0 and 100
        if self.properties['display'] == 'interlaced':
            ypos = self.properties['interval_height'] if self.counter % 2 == 0 else 1

        elif self.properties['display'] == 'collapsed':
            ypos = 0

        else:
            ypos = free_row * self.row_scale
        return ypos
    
    
    def plot(self, ax, chrom_region, start_region, end_region, label_align={}):
        self.counter = 0
        self.small_relative = 0.004 * (end_region - start_region)
        self._get_length_w(ax.get_figure().get_figwidth(), start_region, end_region)

        genes_overlap = sorted(self.interval_tree[chrom_region][start_region:end_region])

        if self.properties['labels'] == 'auto':
            if len(genes_overlap) > 30:
                # turn labels off when too many intervals are visible.
                self.is_draw_labels = False
            else:
                self.is_draw_labels = True

        max_num_row_local = 1
        max_ypos = 0

        row_last_position = []
        for region in genes_overlap:
            
            self.counter += 1
            bed = region.data

            if self.is_draw_labels:
                num_name_characters = len(bed.name) + 2  # +2 to account for an space before and after the name
                bed_extended_end = int(bed.end + (num_name_characters * self.len_w))
            else:
                bed_extended_end = (bed.end + 2 * self.small_relative)

            # get smallest free row
            if len(row_last_position) == 0:
                free_row = 0
                row_last_position.append(bed_extended_end)
            else:
                # get list of rows that are less than bed.start, then take the min
                idx_list = [idx for idx, value in enumerate(row_last_position) if value < bed.start]
                if len(idx_list):
                    free_row = min(idx_list)
                    row_last_position[free_row] = bed_extended_end
                else:
                    free_row = len(row_last_position)
                    row_last_position.append(bed_extended_end)

            rgb, edgecolor = self.get_rgb_and_edge_color(bed)

            ypos = self._get_y_pos(free_row)

            # do not plot if the maximum interval rows to plot is reached
            if 'gene_rows' in self.properties and free_row >= int(self.properties['gene_rows']):
                continue

            if free_row > max_num_row_local:
                max_num_row_local = free_row
            if ypos > max_ypos:
                max_ypos = ypos

            if self.properties['style'] == 'flybase':
                self._draw_gene_with_introns_flybase_style(ax, bed, ypos, rgb, edgecolor)
            else:
                self._draw_gene_simple(ax, bed, ypos, rgb, edgecolor)

            if not self.is_draw_labels:
                pass
            elif bed.end > start_region and bed.start < end_region:
                if bed.name in label_align:
                    if label_align[bed.name] == 'left':
                        ax.text(bed.end + self.small_relative, ypos + (float(self.properties['interval_height']) / 2),
                                bed.name, horizontalalignment='left',
                                verticalalignment='center', fontproperties=self.fp)
                    else:
                        ax.text(bed.start - self.small_relative, ypos + (float(self.properties['interval_height']) / 2),
                                bed.name, horizontalalignment='right',
                                verticalalignment='center', fontproperties=self.fp)
                else:
                    ax.text(bed.end + self.small_relative, ypos + (float(self.properties['interval_height']) / 2),
                                bed.name, horizontalalignment='left',
                                verticalalignment='center', fontproperties=self.fp)

        if self.counter == 0:
            log.warning("*Warning* No intervals were found for file {} "
                        "in Track '{}' for the interval plotted ({}:{}-{}).\n".
                        format(self.properties['file'], self.properties['name'], chrom_region, start_region, end_region))

        ymax = 0

        if 'global_max_row' in self.properties and self.properties['global_max_row'] == 'yes':
            ymin = self.max_num_row[chrom_region] * self.row_scale

        elif 'gene_rows' in self.properties:
            ymin = int(self.properties['gene_rows']) * self.row_scale
        else:
            ymin = max_ypos + self.properties['interval_height']

        log.debug("ylim {},{}".format(ymin, ymax))
        # the axis is inverted (thus, ymax < ymin)
        ax.set_ylim(ymin, ymax)

        if 'display' in self.properties:
            if self.properties['display'] == 'domain':
                ax.set_ylim(-5, 205)
            elif self.properties['display'] == 'collapsed':
                ax.set_ylim(-5, 105)

        ax.set_xlim(start_region, end_region)

        self.plot_label()
    
    def _draw_gene_simple(self, ax, bed, ypos, rgb, edgecolor):
        """
        draws an interval with direction (if given)
        """
        from matplotlib.patches import Polygon

        if bed.strand not in ['+', '-']:
            ax.add_patch(Rectangle((bed.start, ypos), bed.end - bed.start, self.properties['interval_height'],
                                   edgecolor=edgecolor, facecolor=rgb, linewidth=0.5))
        else:
            vertices = self._draw_arrow(ax, bed.start, bed.end, bed.strand, ypos)
            ax.add_patch(Polygon(vertices, closed=True, fill=True,
                                 edgecolor=edgecolor,
                                 facecolor=rgb,
                                 linewidth=0.5))
    
    def _draw_arrow(self, ax, start, end, strand, ypos):

        half_height = float(self.properties['interval_height']) / 2
        if strand == '+':
            x0 = start
            x1 = end  # - self.small_relative
            y0 = ypos
            y1 = ypos + self.properties['interval_height']

            """
            The vertices correspond to 5 points along the path of a form like the following,
            starting in the lower left corner and progressing in a clock wise manner.
            -----------------\
            ---------------- /
            """

            vertices = [(x0, y0), (x0, y1), (x1, y1), (x1 + self.small_relative, y0 + half_height), (x1, y0)]

        else:
            x0 = start  # + self.small_relative
            x1 = end
            y0 = ypos
            y1 = ypos + self.properties['interval_height']

            """
            The vertices correspond to 5 points along the path of a form like the following,
            starting in the lower left corner and progressing in a clock wise manner.
            /-----------------
            \-----------------
            """

            vertices = [(x0, y0), (x0 - self.small_relative, y0 + half_height), (x0, y1), (x1, y1), (x1, y0)]

        return vertices
    
    def _draw_gene_with_introns_flybase_style(self, ax, bed, ypos, rgb, edgecolor):
        """
        draws a gene using different styles
        """
        from matplotlib.patches import Polygon
        if bed.block_count == 0 and bed.thick_start == bed.start and bed.thick_end == bed.end:
            self._draw_gene_simple(ax, bed, ypos, rgb, edgecolor)
            return
        half_height = float(self.properties['interval_height']) / 2
        # draw 'backbone', a line from the start until the end of the gene
        ax.plot([bed.start, bed.end], [ypos + half_height, ypos + half_height], 'black', linewidth=0.5, zorder=-1)

        # get start, end of all the blocks
        positions = []
        for idx in range(0, bed.block_count):
            x0 = bed.start + bed.block_starts[idx]
            x1 = x0 + bed.block_sizes[idx]
            if x0 < bed.thick_start < x1:
                positions.append((x0, bed.thick_start, 'UTR'))
                positions.append((bed.thick_start, x1, 'coding'))

            elif x0 < bed.thick_end < x1:
                positions.append((x0, bed.thick_end, 'coding'))
                positions.append((bed.thick_end, x1, 'UTR'))

            else:
                if x1 < bed.thick_start or x0 > bed.thick_end:
                    type = 'UTR'
                else:
                    type = 'coding'

                positions.append((x0, x1, type))

        # plot all blocks as rectangles except the last if the strand is + or
        # the first is the strand is -, which are drawn as arrows.
        if bed.strand == '-':
            positions = positions[::-1]

        first_pos = positions.pop()
        if first_pos[2] == 'UTR':
            _rgb = 'grey'
        else:
            _rgb = rgb

        vertices = self._draw_arrow(ax, first_pos[0], first_pos[1], bed.strand, ypos)

        ax.add_patch(Polygon(vertices, closed=True, fill=True,
                             edgecolor=edgecolor,
                             facecolor=_rgb,
                             linewidth=0.5))

        for start_pos, end_pos, _type in positions:
            if _type == 'UTR':
                _rgb = 'grey'
            else:
                _rgb = rgb
            vertices = [(start_pos, ypos), (start_pos, ypos + self.properties['interval_height']),
                        (end_pos, ypos + self.properties['interval_height']), (end_pos, ypos)]

            ax.add_patch(Polygon(vertices, closed=True, fill=True,
                                 edgecolor=edgecolor,
                                 facecolor=_rgb,
                                 linewidth=0.5))

    

    


#cython: language_level=3
#cython: boundscheck=False
#cython: cdivision=True

class Loops(object):

    def __init__(self, neoloop_fil):
        
        self.loops = self.parse_loops(neoloop_fil)
    
    def parse_loops(self, filename):

        loops = {}
        with open(filename, 'r') as source:
            for line in source:
                p = line.rstrip().split()
                loci1 = (p[0], int(p[1]), int(p[2]))
                loci2 = (p[3], int(p[4]), int(p[5]))
                label = p[-1].split(',')
                IDs = [label[i] for i in range(0, len(label), 3)]
                neoloop = [int(label[i+2]) for i in range(0, len(label), 3)]
                for i, n in zip(IDs, neoloop):
                    if not i in loops:
                        loops[i] = []
                    loops[i].append((loci1, loci2, n))
        
        return loops



#cython: language_level=3
#cython: boundscheck=False
#cython: cdivision=True

import cooler, logging
from scipy.sparse import triu
from tadlib.hitad.chromLev import Chrom
from pomegranate import NormalDistribution, HiddenMarkovModel, GeneralMixtureModel, State

log = logging.getLogger(__name__)

class Genome(object):

    def __init__(self, uri, balance_type='sweight', cpu_core=1):

        if not balance_type in ['weight', 'sweight']:
            correct = False
        else:
            correct = balance_type
        
        lib = cooler.Cooler(uri)
        res = lib.binsize
        seqs = []
        log.debug('Calculating DIs for each chromosome ...')
        for c in lib.chromnames:
            if not c in ['chr1', 'chr2', 'chr3']:
                continue
            log.debug('Chrom {0} ...'.format(c))
            tdata = triu(lib.matrix(balance=correct, sparse=True).fetch(c)).tocsr()
            work = Chrom(c, res, tdata, 'pseudo', maxapart=3000000)
            work.minWindows(0, work.chromLen, work._dw)
            work.calDI(work.windows, 0)
            work.splitChrom(work.DIs)
            for r in work.regionDIs:
                withzeros = work.regionDIs[r]
                nozeros = withzeros[withzeros!=0]
                if nozeros.size > 20:
                    seqs.append(nozeros)
        
        self.training_data = seqs
        log.debug('Initialize a Hidden Markov Model ...')
        self.model = self.oriHMMParams()
        log.debug('Learning parameters ...')
        self.fit(cpu_core=cpu_core)
    
    def oriHMMParams(self):
        """
        Set initial parameters for the Hidden Markov Model (HMM).
        
        Attributes
        ----------
        HMMParams : dict
            Has 3 keys: "A", state transition matrix, "B" (emission probabilities),
            specifying parameters (Means, Variances, Weights) of the mixture
            Gaussian distributions for each hidden state, and "pi", indicating
            the hidden state weights. This dict will be updated after learning
            procedure.
        """
        hmm = HiddenMarkovModel()
        # GMM emissions
        # 4 Hidden States:
        # 0--start, 1--downstream, 2--upstream, 3--end
        numdists = 3 # Three-distribution Gaussian Mixtures
        var = 7.5 / (numdists - 1)
        means = [[], [], [], []]
        for i in range(numdists):
            means[3].append(i * 7.5 / ( numdists - 1 ) + 2.5)
            means[2].append(i * 7.5 / ( numdists - 1 ))
            means[1].append(-i * 7.5 / ( numdists - 1 ))
            means[0].append(-i * 7.5 / ( numdists - 1 ) - 2.5)
        states = []
        for i, m in enumerate(means):
            tmp = []
            for j in m:
                tmp.append(NormalDistribution(j, var))
            mixture = GeneralMixtureModel(tmp)
            states.append(State(mixture, name=str(i)))
        hmm.add_states(*tuple(states))

        # Transmission matrix
        #A = [[0., 1., 0., 0.],
        #    [0., 0.5, 0.5, 0.],
        #    [0., 0., 0.5, 0.5],
        #    [1., 0., 0., 0.]]
        hmm.add_transition(states[0], states[1], 1)
        hmm.add_transition(states[1], states[1], 0.5)
        hmm.add_transition(states[1], states[2], 0.5)
        hmm.add_transition(states[2], states[2], 0.5)
        hmm.add_transition(states[2], states[3], 0.5)
        hmm.add_transition(states[3], states[0], 1)

        #pi = [0.2, 0.3, 0.3, 0.2]
        hmm.add_transition(hmm.start, states[0], 1)
        hmm.add_transition(states[3], hmm.end, 1)

        hmm.bake()

        return hmm

    
    def fit(self, cpu_core):

        self.model.fit(self.training_data, algorithm='baum-welch', max_iterations=10000,
                  stop_threshold=1e-5, n_jobs=cpu_core, verbose=False)
        

